#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from typing import Any, Callable, Optional, TypeVar, Union
import warnings

import casadi as ca
import numpy as np
import pandas as pd

from ..base import LearningBase
from ....plugins.plugins import LearningManager, LearningVisualizationManager, check_version
from ....util.data import DataSet
from ....util.machine_learning import net_to_casadi_graph
from ....util.util import is_list_like


ML = TypeVar('ML', bound=LearningBase)


class ArtificialNeuralNetwork(LearningBase):
    """Artificial neural network class"""
    def __init__(self, features, labels, id=None, name=None, **kwargs):
        """Constructor method"""
        super().__init__(features, labels, id=id, name=name)

        self._layers = []
        self._weights = None
        self._bias = None
        self._dropouts = None
        self._hidden = None

        self._seed = kwargs.get('seed')

        learning_rate = kwargs.get('learning_rate')
        if learning_rate is None:
            learning_rate = .001
        self._learning_rate = learning_rate

        loss = kwargs.get('loss')
        if loss is None:
            loss = 'mse'
        self._loss = loss

        optimizer = kwargs.get('optimizer')
        if optimizer is None:
            optimizer = 'adam'
        self._optimizer = optimizer

        metric = kwargs.get('metric')
        if metric is None:
            metric = []
        self._metric = metric

        self._data_sets = []

        train_backend = kwargs.get('backend')
        if train_backend is None:
            train_backend = 'pytorch'
        self._set_backend(train_backend)
        self._net = None

        self._check_data_sets()

        self._scaler_x = None
        self._scaler_y = None
        self._train_data = (None, None)
        self._validate_data = (None, None)
        self._test_data = (None, None)
        self._pandas_version_checked = False

    @staticmethod
    def _set_scaling(scaler: Union[str, Callable], backend: Optional[str] = None) -> Any:
        """

        :param scaler:
        :param backend:
        :return:
        """
        # TODO: Can we use something other than Any for typing the returned value?
        if isinstance(scaler, str):
            if backend is not None:
                scaler = LearningManager(backend).setup(scaler)
            else:
                warnings.warn("Scaling was selected, but no scaler backend was supplied. Please select a scaler backend"
                              " or supply your own scaler object. No scaling applied.")
                return None
        if callable(scaler):
            scaler = scaler()
        has_fit = hasattr(scaler, 'fit') and callable(scaler.fit)
        if not has_fit:
            raise RuntimeError("Supplied scaler is missing the method 'fit'")  # TODO: Return a little more info
        has_transform = hasattr(scaler, 'transform') and callable(scaler.transform)
        if not has_transform:
            raise RuntimeError("Supplied scaler is missing the method 'transform'")  # TODO: Return a little more info

        return scaler

    @staticmethod
    def _check_scaler(scaler: Any) -> None:
        """

        :param scaler:
        :return:
        """
        # TODO: See self._set_scaling (regarding Any)
        if not hasattr(scaler, 'mean_'):
            # TODO: Return a little more info
            raise RuntimeError("Supplied scaler is missing the attribute 'mean_'")
        if not hasattr(scaler, 'scale_'):
            # TODO: Return a little more info
            raise RuntimeError("Supplied scaler is missing the attribute 'scale_'")

    def _check_data_sets(self, data_sets=None):
        """

        :param data_sets:
        :return:
        """
        if data_sets is None:
            data_sets = self._data_sets

        if not is_list_like(data_sets):
            data_sets = [data_sets]

        for data_set in data_sets:
            if isinstance(data_set, DataSet):
                for feature in self._features:
                    if feature not in data_set.features:
                        raise ValueError(f"Feature {feature} does not exist in the supplied data set")
                for label in self._labels:
                    if label not in data_set.labels:
                        raise ValueError(f"Label {label} does not exist in the supplied data set")
            else:
                # NOTE: Right now only pandas dataframes are supported
                if not self._pandas_version_checked:
                    check_version('pandas')
                    self._pandas_version_checked = True
                for feature in self._features:
                    if feature not in data_set.columns:
                        raise ValueError(f"Feature {feature} does not exist in the supplied data set")
                for label in self._labels:
                    if label not in data_set.columns:
                        raise ValueError(f"Label {label} does not exist in the supplied data set")

    @LearningBase.features.setter
    def features(self, arg):
        self._features = arg
        self._n_features = len(arg)
        self._check_data_sets()

    @LearningBase.labels.setter
    def labels(self, arg):
        self._labels = arg
        self._n_labels = len(arg)
        self._check_data_sets()

    @property
    def seed(self):
        """

        :return:
        """
        return self._seed

    @seed.setter
    def seed(self, seed):
        self._seed = seed

    @property
    def learning_rate(self):
        """

        :return:
        """
        return self._learning_rate

    @learning_rate.setter
    def learning_rate(self, lr):
        self._learning_rate = lr

    @property
    def loss(self):
        """

        :return:
        """
        return self._loss

    @loss.setter
    def loss(self, loss):
        self._loss = loss

    @property
    def optimizer(self):
        """

        :return:
        """
        return self._optimizer

    @optimizer.setter
    def optimizer(self, opti):
        self._optimizer = opti

    @property
    def metric(self):
        """

        :return:
        """
        return self._metric

    @metric.setter
    def metric(self, metric):
        self._metric = metric

    @property
    def depth(self):
        """

        :return:
        """
        return len([layer for layer in self._layers if layer.type.lower() != 'dropout'])

    @property
    def shape(self):
        """

        :return:
        """
        return (self._n_features,) + tuple(layer.nodes for layer in self._layers if layer.type.lower() != 'dropout') + (
            self._n_labels,)

    @property
    def backend(self):
        """

        :return:
        """
        return self._backend.backend

    @backend.setter
    def backend(self, backend):
        self._set_backend(backend)

    def add_layers(self, layers):
        """

        :param layers:
        :return:
        """
        if isinstance(layers, (list, tuple, set)):
            for layer in layers:
                self.add_layers(layer)
        else:
            self._layers.append(layers)
            layers.parent = self

    def add_data_set(self, data_set):
        """

        :param data_set:
        :return:
        """
        self._check_data_sets(data_set)
        self._data_sets.append(data_set)

    def build_graph(self, weights=None, bias=None):
        """

        :param weights:
        :param bias:
        :return:
        """
        x = ca.SX.sym('x', self._n_features)
        if weights is None and bias is None:
            self._function = self._net.build_graph(x, self._layers, input_scaling=self._scaler_x,
                                                   output_scaling=self._scaler_y)
        else:
            self._function = net_to_casadi_graph({'weights': weights, 'bias': bias}, x, self._layers,
                                                 input_scaling=self._scaler_x, output_scaling=self._scaler_y)

    def prepare_data_set(
            self,
            train_split: float = 1.,
            validation_split: float = 0.,
            scale_data: bool = False,
            scaler: Optional[str] = None,
            scaler_backend: Optional[str] = None,
            shuffle: bool = True
    ) -> None:
        """

        :param train_split:
        :param validation_split:
        :param scale_data:
        :param scaler:
        :param scaler_backend:
        :param shuffle:
        :return:
        """
        data = pd.concat([self._data_sets[0]] + self._data_sets[1:], ignore_index=True, sort=False)
        if isinstance(data, DataSet):
            x_data, y_data = data.raw_data
            # TODO: Make this the default in the future, so we don't have to transpose (therefore transpose for pandas
            #  objects)
            x_data = x_data.T
            y_data = y_data.T
            # NOTE: I don't think at the moment that a length check is necessary here
        else:
            # NOTE: pandas is assumed here for now
            x_data = data[self._features].values
            y_data = data[self._labels].values

            if len(x_data) != len(y_data):
                raise ValueError(f"Dimension mismatch. Features have {len(x_data)} entries and labels have "
                                 f"{len(y_data)} entries.")

        if scale_data or scaler is not None or scaler_backend is not None:
            if scaler is None:
                scaler = 'StandardScaler'
            self.set_scaling(scaler, backend=scaler_backend)

        if self._scaler_x is not None:
            self._scaler_x.fit(x_data)
            self._check_scaler(self._scaler_x)
            x_data = self._scaler_x.transform(x_data)
        if self._scaler_y is not None:
            self._scaler_y.fit(y_data)
            self._check_scaler(self._scaler_y)
            y_data = self._scaler_y.transform(y_data)

        # TODO: Add support for data split directly in DataSet class
        data_set_size = len(data)
        indices = list(range(data_set_size))
        if shuffle:
            np.random.seed(self._seed)
            np.random.shuffle(indices)
        # NOTE: For odd number of data sets this can lead to different sizes although same split percentage was supplied
        train, validate, test = np.split(indices, [int(train_split * data_set_size),
                                                   int((train_split + validation_split) * data_set_size)])

        x_train = x_data[train, :]
        y_train = y_data[train, :]
        self._train_data = (x_train, y_train)

        x_validate = x_data[validate, :]
        y_validate = y_data[validate, :]
        self._validate_data = (x_validate, y_validate)

        x_test = x_data[test, :]
        y_test = y_data[test, :]
        self._test_data = (x_test, y_test)

    def set_input_scaling(self, scaler: Union[str, Callable], backend: Optional[str] = None) -> None:
        """

        :param scaler:
        :param backend:
        :return:
        """
        self._scaler_x = self._set_scaling(scaler, backend=backend)

    def set_output_scaling(self, scaler: str, backend: Optional[str] = None) -> None:
        """

        :param scaler:
        :param backend:
        :return:
        """
        self._scaler_y = self._set_scaling(scaler, backend=backend)

    def set_scaling(self, scaler: Union[str, Callable], backend: Optional[str] = None) -> None:
        """

        :param scaler:
        :param backend:
        :return:
        """
        if isinstance(scaler, str) and backend is not None:
            scaler = LearningManager(backend).setup(scaler)
        self.set_input_scaling(scaler)
        self.set_output_scaling(scaler)

    def setup(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        loss = kwargs.get('loss')
        if loss is None:
            loss = self._loss
        else:
            self._loss = loss

        optimizer = kwargs.get('optimizer')
        if optimizer is None:
            optimizer = self._optimizer
        else:
            self._optimizer = optimizer

        metric = kwargs.get('metric')
        if metric is None:
            metric = self._metric
        else:
            self._metric = metric

        show_tensorboard = kwargs.get('show_tensorboard')
        if show_tensorboard is None:
            show_tensorboard = False
            save_tensorboard = kwargs.get('save_tensorboard')
            if save_tensorboard is None:
                save_tensorboard = False
        else:
            if show_tensorboard:
                save_tensorboard = True
            else:
                save_tensorboard = kwargs.get('save_tensorboard')
                if save_tensorboard is None:
                    save_tensorboard = False
        tensorboard_log_dir = kwargs.get('tensorboard_log_dir')
        if save_tensorboard:
            tensorboard = LearningVisualizationManager('tensorboard').setup(log_dir=tensorboard_log_dir)
        else:
            tensorboard = None
        if show_tensorboard:
            browser = kwargs.get('browser')
            if browser is None:
                browser = 'chrome'
        else:
            browser = None

        device = kwargs.get('device')

        properties = [self._n_features, self._n_labels, self._layers]
        options = {
            'seed': self._seed,
            'learning_rate': self._learning_rate,
            'loss': loss,
            'optimizer': optimizer,
            'metric': metric,
            'tensorboard': tensorboard,
            'browser': browser,
            'device': device
        }

        self._net = self._backend.setup('MLP', *properties, **options)

    def is_setup(self):
        """

        :return:
        """
        if self._net is not None:
            return True
        else:
            return False

    def train(
            self,
            batch_size: int,
            epochs: int,
            verbose: int = 1,
            validation_split: float = 0.,
            test_split: float = 0.,
            scale_data: bool = False,
            scaler: Optional[str] = None,
            scaler_backend: Optional[str] = None,
            shuffle: bool = True,
            patience: Optional[int] = None
    ) -> None:
        """

        :param batch_size:
        :param epochs:
        :param verbose:
        :param validation_split:
        :param test_split:
        :param scale_data:
        :param scaler:
        :param scaler_backend:
        :param shuffle:
        :param patience:
        :return:
        """
        if not self._data_sets:
            raise RuntimeError("No data set to train on was supplied. Please add a data set by using the 'add_data_set'"
                               " method.")

        if validation_split > 1. or validation_split < 0.:
            raise ValueError("Validation split has to be between 0 and 1")
        if test_split > 1. or test_split < 0.:
            raise ValueError("Test split has to be between 0 and 1")
        train_split = 1. - validation_split - test_split
        if train_split <= 0.:
            raise ValueError("Train split is not big enough. Reduce validation split or test split.")

        self.prepare_data_set(train_split=train_split, validation_split=validation_split, scale_data=scale_data,
                              scaler=scaler, scaler_backend=scaler_backend, shuffle=shuffle)

        self._net.train(self._train_data, self._validate_data, batch_size, epochs, verbose, patience, shuffle)
        if all(data.size > 0 for data in self._test_data):
            print("Evaluate on test data")
            self._net.evaluate(data=self._test_data, batch_size=batch_size, verbose=verbose)

        self.build_graph()

    def is_trained(self):
        """

        :return:
        """
        return super().is_setup()

    def predict(self, X_query):
        """

        :param X_query:
        :return:
        """
        # TODO: We could also add a predict method to the wrapper and just link here, but this would not work with
        #  CasADi variables like this
        return self._function(X_query)[0]

    def show_tensorboard(self, browser=None):
        """

        :param browser:
        :return:
        """
        if self.is_trained():
            if browser is None:
                browser = 'chrome'
            self._net.show_tensorboard(browser.lower())
        else:
            warnings.warn("The artificial neural network has not been trained yet. Aborting...")

    def close_tensorboard(self):
        """

        :return:
        """
        if self.is_trained():
            self._net.close_tensorboard()


__all__ = [
    'ArtificialNeuralNetwork'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from typing import Optional, TypeVar, Union

import casadi as ca
import numpy as np

from .dynamic_model.dynamic_model import Model
from .controller.base import Controller
from .machine_learning.base import LearningBase
from .estimator.base import Estimator


Control = TypeVar('Control', bound=Controller)
ML = TypeVar('ML', bound=LearningBase)
Estimate = TypeVar('Estimate', bound=Estimator)


_step = 0


class SimpleControlLoop:
    """
    This class creates a simple feedback control loop with a plant, controller and an optional observer.

    :param plant:
    :param controller:
    :param observer:
    """
    def __init__(self, plant: Model, controller: Union[Control, ML], observer: Optional[Estimate] = None) -> None:
        """Constructor method"""
        if not plant.is_setup():
            plant.setup()
        self._plant = plant
        self._ind_map_controller = {'x': [], 'u': [], 'y': [], 'z': []}
        if not controller.is_setup():
            controller.setup()
        self._controller = controller
        self._controller_is_mpc = False
        self._controller_is_ocp = False
        self._controller_is_pid = False
        self._controller_is_ann = hasattr(self._controller, 'predict')
        if not self._controller_is_ann:
            self._controller_is_mpc = self._controller.type in ['NMPC', 'LMPC']
            self._controller_is_ocp = self._controller.type == 'OCP'
            if not self._controller_is_mpc:
                if self._controller.type == 'PID':
                    self._controller_is_pid = True

        if self._controller_is_ocp:
            self._u_sequence = None

        if self._controller_is_mpc or self._controller_is_ocp:
            # TODO do the same for the estimator
            name_set = set(self._plant.dynamical_state_names)
            self._ind_map_controller['x'] = [i for i, e in enumerate(self._controller._model_orig.dynamical_state_names)
                                             if
                                             e in name_set]
            name_set = set(self._plant.input_names)
            self._ind_map_controller['u'] = [i for i, e in enumerate(self._controller._model_orig.input_names) if
                                             e in name_set]

            name_set = set(self._plant.measurement_names)
            self._ind_map_controller['y'] = [i for i, e in enumerate(self._controller._model_orig.measurement_names) if
                                             e in name_set]

            name_set = set(self._plant.algebraic_state_names)
            self._ind_map_controller['z'] = [i for i, e in enumerate(self._controller._model_orig.algebraic_state_names)
                                             if
                                             e in name_set]

        if observer is not None:
            if not observer.is_setup():
                observer.setup()
            self._optimization_based_observer = hasattr(observer, 'estimate')
            self._learning_based_observer = hasattr(observer, 'predict')
        self._observer = observer

    def _initialize_solution(self, solution):
        """

        :param solution:
        :return:
        """
        # TODO: Process estimates from the observer and predicted states from MPC
        # Process control bounds
        self._update_bounds(solution, ignore_inputs=True)

    def _finalize_solution(self, solution):
        """

        :param solution:
        :return:
        """
        # TODO: Improve the behavior of the solution object (Series class) for updating values at specified indices
        #  (here the initial values)
        self._update_references(solution, index=1)
        self._update_bounds(solution, ignore_states=True)

    def _update_references(self, solution, ignore_states=False, ignore_inputs=False, index=0):
        """

        :param solution:
        :param ignore_states:
        :param ignore_inputs:
        :param index:
        :return:
        """
        if self._controller_is_mpc:
            x_ref = ca.DM.nan(self._plant.n_x)
            if not ignore_states:
                x_ref[self._ind_map_controller['x']] = self._controller.solution.get_by_id('x_reference')[:, index]

            u_ref = ca.DM.nan(self._plant.n_u)
            if not ignore_inputs:
                u_ref[self._ind_map_controller['u']] = self._controller.solution.get_by_id('u_reference')[:, index]
            kwargs = {
                'x_ref': x_ref,
                'u_ref': u_ref
            }
            solution.update(**kwargs)

    def _update_bounds(self, solution, ignore_states=False, ignore_inputs=False):
        """

        :param solution:
        :param ignore_states:
        :param ignore_inputs:
        :return:
        """
        # NOTE: For the path following case, the controller has more bounds than states of the system, for
        #  this we take only the first self._plant.n_x/self._plant.n_u bounds for states and inputs, respectively.
        if self._controller_is_mpc:
            kwargs = {}
            if not ignore_states:
                x_lb = self._controller._x_lb_orig
                if x_lb:
                    kwargs['x_lb'] = ca.DM.nan(self._plant.n_x)
                    kwargs['x_lb'][self._ind_map_controller['x']] = x_lb
                x_ub = self._controller._x_ub_orig
                if x_ub:
                    kwargs['x_ub'] = ca.DM.nan(self._plant.n_x)
                    kwargs['x_ub'][self._ind_map_controller['x']] = x_ub
            if not ignore_inputs:
                u_lb = self._controller._u_lb_orig
                if u_lb:
                    kwargs['u_lb'] = ca.DM.nan(self._plant.n_u)
                    kwargs['u_lb'][self._ind_map_controller['u']] = u_lb
                u_ub = self._controller._u_ub_orig
                if u_ub:
                    kwargs['u_ub'] = ca.DM.nan(self._plant.n_u)
                    kwargs['u_ub'][self._ind_map_controller['u']] = u_ub
            solution.update(**kwargs)

    @staticmethod
    def _reset_step():
        """

        :return:
        """
        global _step
        _step = 0

    def _live_plot(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        solution = self._plant.solution
        if solution.plot_backend == 'latex':
            raise NotImplementedError("Live animations using the backend 'latex' are not supported")
        elif solution.plot_backend == 'bokeh':
            self._live_plot_bokeh(solution, *args, **kwargs)
        elif solution.plot_backend == 'matplotlib':
            self._live_plot_matplotlib(solution, *args, **kwargs)
        else:
            raise NotImplementedError(f"Plot backend '{solution.plot_device}' not available")

    def _live_plot_bokeh(self, solution, *args, **kwargs):
        """
        Runs a live plot

        :param args:
        :param kwargs:
        :return:
        """
        steps = args[0]
        p = kwargs.get('p')
        # TODO: Check, if we can use solution.get_by_name('t:0') here
        x_range = [float(solution.get_by_name('t')[0]), steps * solution.dt]
        # NOTE: self._plant.n_x * [['line']] will generate a list of self._plant.n_x pointers to one and the same list,
        #  so we use list comprehension here.
        # kind = [['line'] for _ in range(self._plant.n_x)] + \
        #        [['line'] for _ in range(self._plant.n_y)] + \
        #        [['line'] for _ in range(self._plant.n_z)] + \
        #        [['step'] for _ in range(self._plant.n_u)] + \
        #        [['step'] for _ in range(self._plant.n_p)]

        plot, sources, Server, theme = self.plot(server=True, x_range=x_range)  # kind=kind
        browser = kwargs.get('browser')

        # TODO: Improve this (right now it only works if no arguments are supplied, so we plot everything)
        names = solution.get_names()[1:]  # Ignore time 't'

        def bkapp(doc):
            """

            :param doc:
            :return:
            """
            def callback():
                """

                :return:
                """
                # TODO: What are unlocked callbacks? (see https://docs.bokeh.org/en/latest/docs/user_guide/server.html)
                global _step

                x0 = solution.get_by_id('x:f')
                self._run(x0, _step, p=p)
                _step += 1

                # BEST PRACTICE --- update .data in one step with a new dict
                # TODO: Needs to be updated according to plot specifications.
                # NOTE: This workflow assumes, that we only have one variable per subplot and the first entry of the
                #  list 'source' is always the solution and the second entry is always the reference. This will
                #  obviously not work with more complicated plot arrangements the way it is implemented right now.
                #  Needs to be updated in the future.
                for k, source in enumerate(sources):
                    x = solution.get_by_name('t').full().flatten()
                    y = solution.get_by_name(names[k]).full().flatten()
                    if x.size > y.size:  # in case of inputs -> step plot
                        x = x[:-1]
                    new_data = {
                        'x': x,
                        'y0': y
                    }
                    source[0].data = new_data  # In the current setup first source is always the data

                    # reference = solution.get_ref_by_name(names[k])
                    # if reference is not None and not reference.is_empty():
                    #     new_data['y0'] = reference.full().flatten()
                    #     source[1].data = new_data  # In the current setup second source is always the reference

                if _step >= steps:
                    doc.remove_periodic_callback(pc_id)

            button = plot.children[0]
            # button.label = "Start live animation"
            # button.on_click(callback)
            button.visible = False

            doc.add_root(plot)
            pc_id = doc.add_periodic_callback(callback, 200)

            # doc.theme = theme

        server = Server({'/': bkapp})  # num_procs=4
        server.start()
        server.io_loop.add_callback(server.show, "/", browser=browser)
        server.io_loop.start()
        # TODO: How to shutdown server

    def _live_plot_matplotlib(self, solution, *args, **kwargs):
        """

        :param solution:
        :param args:
        :param kwargs:
        :return:
        """
        steps = args[0]
        iterations = args[0]
        p = kwargs.get('p')
        # TODO: Check, if we can use solution.get_by_name('t:0') here
        x_range = [float(solution.get_by_name('t')[0]), steps * solution.dt]

        plt, fig, axes, anim = self.plot(anim=True, x_range=x_range)
        axes = axes.flatten()
        lines = [ax.get_lines() for ax in axes]

        # TODO: Improve this (right now it only works if no arguments are supplied, so we plot everything)
        names = solution.get_names()[1:]  # Ignore time 't'

        def update(frame):
            """

            :param frame:
            :return:
            """
            nonlocal steps

            if steps > 0:
                x0 = solution.get_by_id('x:f')
                self._run(x0, iterations - steps, p=p)

                for k, line in enumerate(lines):
                    x = solution.get_by_name('t').full().flatten()
                    y = solution.get_by_name(names[k]).full().flatten()
                    if x.size > y.size:  # in case of inputs -> step plot
                        x = x[:-1]
                    line[0].set_data(x, y)  # In the current setup first line is always the data

                    # reference = solution.get_ref_by_name(names[k])
                    # if reference is not None and not reference.is_empty():
                    #     ref = reference.full().flatten()
                    #     line[1].set_data(x, ref)  # In the current setup second line is always the reference

                    axes[k].relim()
                    axes[k].autoscale_view()

                steps -= 1

            return np.concatenate(lines)

        # NOTE: For whatever reason just writing anim(...) won't work
        ani = anim(fig, update, frames=np.arange(steps), blit=True)
        plt.show()

    def _run(self, x0, iteration, p=None, **kwargs):
        """

        :param x0:
        :param iteration:
        :param p:
        :param kwargs:
        :return:
        """
        # Controller step
        if self._controller_is_mpc or self._controller_is_ocp:
            # states of model could be different from the states of the plant.
            state_names = self._controller._model_orig.dynamical_state_names
            ind_states = [self._plant.dynamical_state_names.index(name) for name in state_names]

            if len(self._controller._time_varying_parameters) > 0:
                ind_param = [name for name in self._plant.parameter_names if
                             name not in self._controller._time_varying_parameters]
                cp = [p[i] for i in ind_param]
            else:
                cp = p

            if self._controller_is_mpc:
                u = self._controller.optimize(x0[ind_states], cp=cp, **kwargs)
            else:
                if iteration == 0:
                    _ = self._controller.optimize(x0[ind_states], cp=cp, **kwargs)
                    self._u_sequence = self._controller.solution['u']
                u = self._u_sequence[:, iteration]
        elif self._controller_is_ann:
            u = self._controller.predict(x0)
        elif self._controller_is_pid:
            u = self._controller.call(pv=x0)
        else:
            u = self._controller.call(x=x0, p=p)

        # Update reference
        # TODO: Improve processing of references (something similar to get_function_args for references?)
        # TODO: References supplied by user
        self._update_references(self._plant.solution)
        self._update_bounds(self._plant.solution)

        # Simulate plant
        self._plant.simulate(u=u, p=p)

        # Observer step
        if self._observer is not None:
            if self._optimization_based_observer:
                # TODO: Further processing
                # It's an MHE or KF
                self._observer.estimate()
            elif self._learning_based_observer:
                # TODO: Further processing
                # It's a learning-based 'observer'
                self._observer.predict()

    def run(self, steps, p=None, live_animation=False, browser=None, **kwargs):
        """

        :param steps: Number of simulation steps
        :type steps: int
        :param p:
        :type p:
        :param live_animation:
        :type live_animation: bool
        :param browser:
        :param kwargs:
        :return:
        """
        if self._controller_is_ocp:
            if not self._controller.prediction_horizon >= steps:
                raise ValueError(f"To solve an OCP problem, the prediction horizon must be larger than the number of "
                                 f"steps. The horizon length is {self._controller.prediction_horizon}, while the steps"
                                 f" are {steps}.")

        if _step != 0:
            self._reset_step()

        if live_animation:
            self._live_plot(steps, p=p, browser=browser)
        else:
            self._plant.reset_solution()
            solution = self._plant.solution
            self._initialize_solution(solution)
            x0 = solution.get_by_id('x:0')
            for k in range(steps):
                self._run(x0, k, p=p, **kwargs)
                x0 = solution.get_by_id('x:f')
            self._finalize_solution(solution)

    def plot(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        new_kwargs = kwargs
        if self._plant.solution.plot_backend == 'bokeh':
            if kwargs.get("output_notebook", False):
                new_kwargs["figsize"] = kwargs.get("figsize", (300, 300))
            else:
                new_kwargs["figsize"] = kwargs.get("figsize", (400, 400))
            new_kwargs["line_width"] = kwargs.get("line_width", 2)
        elif self._plant.solution.plot_backend == 'matplotlib':
            new_kwargs["line_width"] = kwargs.get("line_width", 1)

        # NOTE: The following 2 lines will be ignored for backend 'matplotlib'
        new_kwargs["major_label_text_font_size"] = kwargs.get("major_label_text_font_size", "12pt")
        new_kwargs["axis_label_text_font_size"] = kwargs.get("axis_label_text_font_size", "12pt")
        new_kwargs["layout"] = kwargs.get("layout", (-1, 3))

        return self._plant.solution.plot(*args, **new_kwargs)



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional, Sequence, Tuple, TypeVar, Union, Iterator
import warnings

import casadi as ca
import numpy as np
from scipy import stats

from .inference import Inference
from .likelihood import Likelihood
from .mean import Mean
from .kernel import Kernel
from ..base import LearningBase
from ...base import Series, TimeSeries, Equations
from ...optimizer import NonlinearProgram
from ....util.machine_learning import Parameter, Hyperparameter, register_hyperparameters
from ....util.util import convert, is_list_like


Numeric = Union[int, float]
Values = Union[Numeric, np.ndarray]
Bound = Union[Numeric, str, Tuple[Numeric], Tuple[Numeric, Numeric]]
Bounds = Dict[str, Union[str, Tuple[Numeric, Numeric]]]
Array = TypeVar('Array', np.ndarray, ca.DM, ca.SX, ca.MX)
Param = TypeVar('Param', bound=Parameter)
Inf = TypeVar('Inf', bound=Inference)
Lik = TypeVar('Lik', bound=Likelihood)
Mu = TypeVar('Mu', bound=Mean)
Cov = TypeVar('Cov', bound=Kernel)


@dataclass
class Data:
    """"""
    values: np.ndarray = field(default_factory=lambda: np.array([[]]))
    SX: ca.SX = field(default_factory=ca.SX)
    MX: ca.MX = field(default_factory=ca.MX)


class _GPSeries(Series):
    """"""
    def __init__(self, backend: str) -> None:
        super().__init__(backend=backend)

        self._abscissa = 'x'

    def _update_dimensions(self) -> None:
        """

        :return:
        """
        pass


class GaussianProcess(LearningBase):
    """
    Gaussian Process Regression

    :Note: The Gaussian process regressor currently does not use the Cholesky factorization for training and
        prediction. This will be implemented at a later time.

    :param features: names of the features
    :type features: list of strings
    :param labels: names of the labels
    :type labels: list of strings
    :param inference:
    :type inference:
    :param likelihood:
    :type likelihood:
    :param kernel: Specifying the covariance function of the GP. The kernel hyperparameters are optimized during
        training, defaults to :class:`hilo_mpc.core.learning.kernels.SquaredExponential`.
    :type kernel: kernel
    :param noise_variance:
    :type noise_variance:
    :param hyperprior:
    :type hyperprior:
    :param id:
    :type id: str, optional
    :param name: name of the GP
    :type name: str, optional
    :param solver:
    :type solver: str, optional
    :param solver_options: Options to the solver, e.g. ipopt.
    :type solver_options: dict
    :param kwargs:
    """
    def __init__(
            self,
            features: Union[str, list[str]],
            labels: Union[str, list[str]],
            inference: Optional[Union[str, Inf]] = None,
            likelihood: Optional[Union[str, Lik]] = None,
            mean: Optional[Mu] = None,
            kernel: Optional[Cov] = None,
            noise_variance: Numeric = 1.,
            hyperprior: Optional[str] = None,  # Maybe we should be able to access other hyperpriors from here as well
            id: Optional[str] = None,
            name: Optional[str] = None,
            solver: Optional[str] = None,
            solver_options: Optional[dict] = None,  # TODO: Switch to mapping?
            **kwargs
    ) -> None:
        """Constructor method"""
        if not is_list_like(features):
            features = [features]
        if not is_list_like(labels):
            labels = [labels]
        if len(labels) > 1:
            raise ValueError("Training a GP on multiple labels is not supported. Please use 'MultiOutputGP' to train "
                             "GPs on multiple labels.")
        super().__init__(features, labels, id=id, name=name)

        if likelihood is None:
            likelihood = Likelihood.gaussian()
        elif isinstance(likelihood, str):
            name = likelihood.replace("'", "").replace(' ', '_').lower()
            if name == 'gaussian':
                likelihood = Likelihood.gaussian()
            elif name == 'logistic':
                likelihood = Likelihood.logistic()
            elif name == 'laplacian':
                likelihood = Likelihood.laplacian()
            elif name == 'students_t':
                likelihood = Likelihood.students_t()
            else:
                raise ValueError(f"Likelihood '{likelihood}' not recognized")
        self.likelihood = likelihood

        if inference is None:
            inference = Inference.exact()
        elif isinstance(inference, str):
            name = inference.replace(' ', '_').lower()
            if name == 'exact':
                inference = Inference.exact()
            elif name == 'laplace':
                inference = Inference.laplace()
            elif name == 'expectation_propagation':
                inference = Inference.expectation_propagation()
            elif name == 'variational_bayes':
                inference = Inference.variational_bayes()
            elif name == 'kullback_leibler':
                inference = Inference.kullback_leibler()
            else:
                raise ValueError(f"Inference '{inference}' not recognized")
        self.inference = inference

        if mean is None:
            mean = Mean.zero()
        self.mean = mean

        if kernel is None:
            kernel = Kernel.squared_exponential()
        self.kernel = kernel

        hyper_kwargs = {}
        if hyperprior is not None:
            hyper_kwargs['prior'] = hyperprior
        hyperprior_parameters = kwargs.get('hyperprior_parameters')
        if hyperprior_parameters is not None:
            hyper_kwargs['prior_parameters'] = hyperprior_parameters
        bounds = kwargs.get('bounds')
        if bounds is not None:  # pragma: no cover
            variance_bounds = bounds.get('noise_variance')
            if variance_bounds is not None:
                if variance_bounds == 'fixed':
                    hyper_kwargs['fixed'] = True
                else:
                    hyper_kwargs['bounds'] = bounds
        self.noise_variance = Hyperparameter('GP.noise_variance', value=noise_variance, **hyper_kwargs)

        self._hyp_is_log = {'GP.noise_variance': True}
        self._hyp_is_log.update({parameter.name: False for parameter in self.mean.hyperparameters})
        self._hyp_is_log.update({parameter.name: True for parameter in self.kernel.hyperparameters})

        register_hyperparameters(self, [parameter.id for parameter in self.hyperparameters])

        # bounds = self.noise_variance.log_bounds
        # unbounded_noise = bounds[0] == -ca.inf and bounds[1] == ca.inf
        # unconstrained_op = all([not self.kernel.is_bounded(), unbounded_noise])
        unconstrained_op = True

        if solver is None:
            if unconstrained_op:
                solver = 'Newton-CG'
            else:  # pragma: no cover
                # TODO: Test GP for constrained hyperparameters
                solver = 'L-BFGS-B'
        self._solver = solver

        if solver == 'ipopt':
            if solver_options is None:
                solver_options = {'print_time': False, 'ipopt.suppress_all_output': 'yes'}
            # solver_options.update({'ipopt.hessian_approximation': 'limited-memory'})
        else:
            if solver_options is None:
                solver_options = {}
        self._solver_options = solver_options

        self._X_train = Data()
        self._y_train = Data()

        epsilon = kwargs.get('epsilon')
        if epsilon is None:
            epsilon = 1e-8
        self._epsilon = epsilon

        self._where_is_what = {}
        self._log_marginal_likelihood = None
        self._gp_solver = None
        self._gp_args = {}
        self._optimization_stats = {}

    def __str__(self) -> str:
        """String representation method"""
        message = "Gaussian process with \n"
        for attribute, value in self.__dict__.items():
            if attribute[0] != '_':
                message += f"\t {attribute}: {value} \n"
        return message

    def _initialize_solver(self) -> None:
        """

        :return:
        """
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            self._gp_solver = NonlinearProgram(solver=self._solver, solver_options=self._solver_options)
        for warning in w:
            if issubclass(warning.category, UserWarning):
                if warning.message.args[0] == 'Plots are disabled, since no backend was selected.':
                    continue
            warnings.warn(warning.message, warning.category)

    def _update_gp_args(
            self,
            name: str,
            value: Values,
            bounds: Optional[Bound]
    ) -> None:
        """

        :param name:
        :param value:
        :param bounds:
        :return:
        """
        if self._gp_solver is not None:
            x_or_p, index = self._where_is_what[name]
            is_slice = isinstance(index, slice)

            if hasattr(value, 'shape'):
                shape = value.shape
                if not shape:
                    n_val = 1
                else:
                    n_dim = len(shape)
                    if n_dim == 1:
                        n_val = shape[0]
                    else:
                        if n_dim == 2:
                            if 1 in shape:
                                n_val = shape[0] * shape[1]
                            else:
                                raise ValueError("Dimension mismatch. Expected 1 dimension, got 2.")
                        else:
                            raise ValueError(f"Dimension mismatch. Expected 1 dimension, got {n_dim}.")
            else:
                n_val = 1
            if (is_slice and n_val != index.stop - index.start) or (is_slice and n_val == 1) or (
                    not is_slice and n_val > 1):
                # TODO: Raise error here?
                msg = f"Dimension of hyperparameter {name} changed. Run the setup() method again"
                warnings.warn(msg)
            else:
                if self._hyp_is_log[name]:
                    if 'variance' in name:
                        self._gp_args[x_or_p][index] = ca.log(value) / 2
                    else:
                        self._gp_args[x_or_p][index] = ca.log(value)
                else:
                    self._gp_args[x_or_p][index] = value

            if bounds is not None:  # pragma: no cover
                if bounds == 'fixed':
                    if n_val > 1:
                        raise ValueError(f"Dimension mismatch between values and bounds. Supplied values have a "
                                         f"dimension bigger than 1.")
                    lb = value
                    ub = value
                elif isinstance(bounds, (int, float)):
                    lb = bounds
                    ub = ca.inf
                elif len(bounds) == 1:
                    lb = bounds[0]
                    ub = ca.inf
                elif len(bounds) == 2:
                    lb = bounds[0]
                    ub = bounds[1]
                    if lb > ub:
                        lb, ub = ub, lb
                else:
                    raise ValueError("Bounds of unsupported type")
                if lb is not None:
                    self._gp_args['lbx'][index] = lb
                if ub is not None:
                    self._gp_args['ubx'][index] = ub

    @property
    def X_train(self) -> Data:
        """
        The tuple contains the training data matrix as ndarray of shape (dimensions_input, number_observations) and
        its respective CasADi SX and MX symbols of same dimensions that are used to define the CasADi functions during
        fitting and prediction. It can be changed by just assigned an ndarray and its setter method will automatically
        adjust the symbols.

        :return:
        """
        return self._X_train

    @X_train.setter
    def X_train(self, value: np.ndarray) -> None:
        shape = value.shape
        if shape[0] != self._n_features:
            raise ValueError(f"Dimension mismatch. Supplied dimension for the features is {shape[0]}, but "
                             f"required dimension is {self._n_features}.")
        X_train_SX = ca.SX.sym('X', *shape)
        X_train_MX = ca.MX.sym('X', *shape)
        self._X_train = Data(values=value, SX=X_train_SX, MX=X_train_MX)

    @property
    def y_train(self) -> Data:
        """
        The tuple contains the training target matrix as ndarray of shape (dimensions_input, number_observations)
        and its respective CasADi SX and MX symbols of same dimensions that are used to define the CasADi functions
        during fitting and prediction. It can be changed by just assigned an ndarray and its setter method will
        automatically adjust the symbols.

        :return:
        """
        return self._y_train

    @y_train.setter
    def y_train(self, value: np.ndarray) -> None:
        shape = value.shape
        if shape[0] != self._n_labels:
            raise ValueError(f"Dimension mismatch. Supplied dimension for the labels is {shape[0]}, but required "
                             f"dimension is {self._n_labels}.")
        y_train_SX = ca.SX.sym('y', *shape)
        y_train_MX = ca.MX.sym('y', *shape)
        self._y_train = Data(values=value, SX=y_train_SX, MX=y_train_MX)

    def set_training_data(self, X: np.ndarray, y: np.ndarray) -> None:
        """Sets the training matrix and its target vector

        :param X:
        :param y:
        :return:
        """
        if self._function is not None:
            warnings.warn("Gaussian process was already executed. Use the fit_model() method again to optimize with "
                          "respect to the newly set training data.")

        new_X_shape = X.shape
        new_y_shape = y.shape
        if new_X_shape[1] != new_y_shape[1]:
            raise ValueError("Number of observations in training matrix and target vector do not match!")
        old_X_shape = self._X_train.values.shape
        old_y_shape = self._y_train.values.shape
        self.X_train = X
        self.y_train = y

        if self._gp_solver is not None:
            if old_X_shape == new_X_shape and old_y_shape == new_y_shape:
                n = new_X_shape[1]
                n *= self._n_features
                self._gp_args['p'][:n] = X.flatten()
                self._gp_args['p'][n:2 * n] = y.flatten()
            else:
                warnings.warn("Dimensions of training data set changed. Please run setup() method again.")

    @property
    def hyperparameters(self) -> list[Param]:
        """
        Includes all hyperparameters of the kernel as well as the sigma_noise hyperparameter

        :return:
        """
        return [self.noise_variance] + self.mean.hyperparameters + self.kernel.hyperparameters

    @property
    def hyperparameter_names(self) -> list[str]:
        """

        :return:
        """
        return [self.noise_variance.name] + self.mean.hyperparameter_names + self.kernel.hyperparameter_names

    def get_hyperparameter_by_name(self, name: str) -> Param:
        """

        :param name:
        :return:
        """
        return [parameter for parameter in self.hyperparameters if parameter.name == name][0]

    def update(self, *args) -> None:
        """

        :param args:
        :return:
        """
        self._update_gp_args(*args)

    def update_hyperparameters(
            self,
            names: Sequence[str],
            values: Optional[Union[Sequence[Optional[Values]], np.ndarray]] = None,
            bounds: Optional[Sequence[Optional[Bounds]]] = None
    ) -> None:
        """
        Updates hyperparameter values and boundaries

        If names of hyperparameters are found in the given dictionaries, their respective value and/or boundaries will
        be updated to the value and/or boundaries defined in the dictionary.

        :Note: Instead of boundary values the string 'fixed' can be passed which flags the hyperparameter, keeping him
            constant during any optimization routine. This flag will be automatically set back if a list of two floats
            is passed.

        :param names:
        :param values:
        :param bounds:
        :return:
        """
        for name in names:
            parameter = self.get_hyperparameter_by_name(name)
            _, index = self._where_is_what[name]
            if values is not None:
                if hasattr(values, 'ndim') and values.ndim != 1:
                    raise ValueError("At the moment only flat NumPy arrays are supported for updating the "
                                     "hyperparameter values.")
                value = values[index]
                if value is not None:
                    if 'variance' in name:
                        value *= 2
                    if not parameter.fixed:
                        if self._hyp_is_log[name]:
                            value = np.exp(value)
                    parameter.value = value
            if bounds is not None:  # pragma: no cover
                bound = bounds[index]
                if bound is not None:
                    parameter.bounds = bound

    def initialize(
            self,
            features: Union[str, list[str]],
            labels: Union[str, list[str]],
            inference: Optional[Union[str, Inf]] = None,
            likelihood: Optional[Union[str, Lik]] = None,
            mean: Optional[Mu] = None,
            kernel: Optional[Cov] = None,
            noise_variance: Numeric = 1.,
            hyperprior: Optional[str] = None,  # Maybe we should be able to access other hyperpriors from here as well
            id: Optional[str] = None,
            name: Optional[str] = None,
            solver: Optional[str] = None,
            solver_options: Optional[dict] = None,  # TODO: Switch to mapping?
            **kwargs
    ) -> None:
        """

        :param features:
        :param labels:
        :param inference:
        :param likelihood:
        :param mean:
        :param kernel:
        :param noise_variance:
        :param hyperprior:
        :param id:
        :param name:
        :param solver:
        :param solver_options:
        :param kwargs:
        :return:
        """
        # NOTE: Maybe we could abuse the __call__-method and check in there whether the class was already instantiated
        #  (https://stackoverflow.com/questions/40457599/how-to-check-if-an-object-has-been-initialized-in-python/40459563,
        #  https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python).
        #  If the object was already instantiated, we execute the normal __call__-method.
        self.__init__(features, labels, inference=inference, likelihood=likelihood, mean=mean, kernel=kernel,
                      noise_variance=noise_variance, hyperprior=hyperprior, id=id, name=name, solver=solver,
                      solver_options=solver_options, **kwargs)

    def setup(self, **kwargs) -> None:
        """

        :param kwargs:
        :return:
        """
        X_sym = self._X_train.SX
        if X_sym.is_empty():
            raise RuntimeError("The training data has not been set. Please run the method set_training_data() to "
                               "proceed.")

        y_sym = self._y_train.SX
        if y_sym.is_empty():
            # NOTE: We shouldn't get here
            raise RuntimeError("The training data has not been set. Please run the method set_training_data() to "
                               "proceed.")

        n, D = X_sym.shape
        X = ca.SX.sym('X', n)

        posterior = self.inference(X_sym, y_sym, X, self.noise_variance.SX, self.likelihood, self.mean, self.kernel)
        log_marginal_likelihood = posterior['log_marginal_likelihood']
        mean = posterior['mean']
        var = posterior['var']

        hyperparameters_to_optimize = [parameter for parameter in self.hyperparameters if not parameter.fixed]
        w = []
        w0 = []
        # lbw = []
        # ubw = []
        k = 0
        for parameter in hyperparameters_to_optimize:
            name = parameter.name
            param = parameter.SX

            hyperprior = parameter.prior
            if hyperprior is not None:
                log_marginal_likelihood += hyperprior(param, log=True)

            w.append(param)
            n_p = param.numel()
            if self._hyp_is_log[name]:
                if 'variance' in name:
                    w0.append(parameter.log / 2)
                else:
                    w0.append(parameter.log)
                # lbw.append(n_p * [parameter.log_bounds[0]])
                # ubw.append(n_p * [parameter.log_bounds[1]])
            else:
                w0.append(parameter.value)
                # lbw.append(n_p * [parameter.bounds[0]])
                # ubw.append(n_p * [parameter.bounds[1]])
            if n_p == 1:
                self._where_is_what[name] = ('x0', k)
            else:
                self._where_is_what[name] = ('x0', slice(k, k + n_p))
            k += n_p

        hyperparameters_fixed = [parameter for parameter in self.hyperparameters if parameter.fixed]
        p = []
        p0 = []
        k = 2 * D
        for parameter in hyperparameters_fixed:
            name = parameter.name
            p.append(parameter.SX)
            n_p = parameter.SX.numel()
            if self._hyp_is_log[name]:
                if 'variance' in name:
                    p0.append(parameter.log / 2)
                else:
                    p0.append(parameter.log)
            else:
                p0.append(parameter.value)
            if n_p == 1:
                self._where_is_what[name] = ('p', k)
            else:
                self._where_is_what[name] = ('p', slice(k, k + n_p))
            k += n_p

        w = ca.vertcat(*w)
        # Flatten all elements in w0 before vertcat (handles both scalars and arrays)
        w0_flat = [np.asarray(val).flatten() for val in w0]
        w0 = ca.vertcat(*w0_flat)
        # lbw = ca.vertcat(*lbw)
        # ubw = ca.vertcat(*ubw)
        p = ca.vertcat(X_sym.T[:], y_sym.T[:], *p)
        # TODO: Check if this is actually the same as what is done for the SX variables
        # Flatten all elements in p0 before concatenate (handles both scalars and arrays)
        p0_flat = [np.asarray(val).flatten() for val in p0]
        p0 = np.concatenate([self.X_train.values.flatten(), self.y_train.values.flatten()] + p0_flat)

        self._initialize_solver()

        # for some reason this doesn't result in free variables
        # self._gp_solver.set_decision_variables(w, lower_bound=lbw, upper_bound=ubw)
        self._gp_solver.set_decision_variables(w)
        self._gp_solver.set_parameters(p)
        self._gp_solver.set_objective(-log_marginal_likelihood)

        self._log_marginal_likelihood = ca.Function('log_marginal_likelihood',
                                                    [w, p],
                                                    [log_marginal_likelihood],
                                                    ['x0', 'p'],
                                                    ['log_marg_lik'])

        self._function = ca.Function(
            'prediction',
            [X, w, p],
            [mean, var],
            ['X', 'x0', 'p'],
            ['mean', 'variance']
        )

        self._gp_solver.setup()

        self._gp_solver.set_initial_guess(w0)
        self._gp_solver.set_parameter_values(p0)

        self._gp_args.update({
            'x0': w0,
            'p': p0,
            # 'lbx': lbw,
            # 'ubx': ubw
        })

    def is_setup(self) -> bool:
        """

        :return:
        """
        if self._gp_solver is not None:
            return True
        else:
            return False

    def log_marginal_likelihood(self) -> float:
        """

        :return:
        """
        return float(self._log_marginal_likelihood(x0=self._gp_args['x0'], p=self._gp_args['p'])['log_marg_lik'])

    def fit_model(self) -> None:
        """
        Optimizes the hyperparameters by minimizing the negative log marginal likelihood.

        The model fit is the training phase in the GP regression. Given the design data (training observations and
        their targets) and suitable start values and bounds for the hyperparameters, the hyperparameters will be
        adapted by minimizing the negative log marginal likelihood.

        :Note: Instead of boundary values the string 'fixed' can be passed which flags the hyperparameter, keeping it
            constant during the optimization routine.

        :return:
        """
        if self._gp_solver is None:
            raise RuntimeError("The GP has not been set up yet. Please run the setup() method before fitting.")

        self._gp_solver.solve()
        solution = self._gp_solver.solution

        names = [parameter.name for parameter in self.hyperparameters if not parameter.fixed]
        values = solution.get_by_id('x:f').full().flatten()
        self.update_hyperparameters(names, values=values)
        self._optimization_stats = self._gp_solver.stats()
        if not self._optimization_stats['success']:  # pragma: no cover
            if self._solver == 'ipopt':
                return_status = self._optimization_stats['return_status']
                if return_status == 'Infeasible_Problem_Detected':
                    message = 'Infeasible problem detected'
                elif return_status == 'Restoration_Failed':
                    message = 'Restoration failed'
                elif return_status == 'Maximum_Iterations_Exceeded':
                    message = 'Maximum iterations exceeded'
                else:
                    raise RuntimeError(f"Unrecognized return status: {return_status}")
            else:
                message = self._optimization_stats['message']
            warnings.warn(f"Fitting of GP didn't terminate successfully\nSolver message: {message}\n"
                          f"Try to use a different solver")

    def predict(self, X_query: Array, noise_free: bool = False) -> Tuple[Array, Array]:
        """

        :param X_query:
        :param noise_free:
        :return:
        """
        if self._function is None:
            raise RuntimeError("The GP has not been set up yet. Please run the setup() method before predicting.")

        prediction = self._function(X=X_query, x0=self._gp_args['x0'], p=self._gp_args['p'])
        mean, var = prediction['mean'], prediction['variance']
        if not noise_free:
            var += self.noise_variance.value

        if isinstance(X_query, np.ndarray):
            mean = mean.full()
            var = var.full()

        return mean, var

    def predict_quantiles(
            self,
            quantiles: Optional[tuple[Numeric, Numeric]] = None,
            X_query: Optional[np.ndarray] = None,
            mean: Optional[np.ndarray] = None,
            var: Optional[np.ndarray] = None
    ) -> Optional[tuple[np.ndarray, np.ndarray]]:
        """

        :param quantiles:
        :param X_query:
        :param mean:
        :param var:
        :return:
        """
        if quantiles is None:
            quantiles = (2.5, 97.5)

        if X_query is not None:
            mean, var = self.predict(X_query, noise_free=True)
        elif mean is None and var is None:
            return None

        predictive_quantiles = [stats.norm.ppf(q / 100) * np.sqrt(var + float(self.noise_variance.value)) + mean for q
                                in quantiles]

        return predictive_quantiles[0], predictive_quantiles[1]

    def plot(self, X_query: Array, backend: Optional[str] = None, **kwargs) -> None:
        """

        :param X_query:
        :param backend:
        :param kwargs:
        :return:
        """
        X_train = self._X_train.values
        y_train = self._y_train.values
        n_x = self._n_features

        post_mean, _ = self.predict(X_query, noise_free=True)

        solution = _GPSeries(backend=backend)
        if backend is None:
            backend = solution.plot_backend

        names = ['x', 'y']
        vector = {
            'x': {
                'values_or_names': self._features,
                'description': n_x * [''],
                'labels': self._features,
                'units': n_x * [''],
                'shape': (n_x, 0),
                'data_format': ca.DM
            },
            'y': {
                'values_or_names': ['post. mean at test points'],
                'description': [''],
                'labels': self._labels,
                'units': [''],
                'shape': (1, 0),
                'data_format': ca.DM
            }
        }
        solution.setup(*names, **vector)

        solution.set('x', X_query)
        solution.set('y', post_mean)

        plots = self._n_labels * tuple((feature, 'post. mean at test points') for feature in self._features)

        ext_data_x = []
        ext_data_y = []
        for i in range(self._n_labels):
            for j in range(n_x):
                ext_data_x.append({'data': X_train[j, :], 'subplot': i * n_x + j, 'label': 'training points'})
                ext_data_y.append(
                    {
                        'data': y_train[i, :],
                        'subplot': i * n_x + j,
                        'label': 'training points',
                        'kind': 'scatter',
                        'marker': '+',
                        'marker_size': 20 if backend == 'bokeh' else 60
                    }
                )

        plot_kwargs = kwargs.copy()
        if backend == 'bokeh':
            if kwargs.get("output_notebook", False):
                plot_kwargs["figsize"] = kwargs.get("figsize", (300, 300))
            else:
                plot_kwargs["figsize"] = kwargs.get("figsize", (500, 500))
            plot_kwargs["major_label_text_font_size"] = kwargs.get("major_label_text_font_size", "12pt")
            plot_kwargs["axis_label_text_font_size"] = kwargs.get("axis_label_text_font_size", "12pt")
        plot_kwargs['marker'] = kwargs.get('marker', 'o')
        if backend == 'bokeh':
            plot_kwargs['marker_size'] = kwargs.get('marker_size', 5)
        elif backend == 'matplotlib':
            plot_kwargs['marker_size'] = kwargs.get('marker_size', 20)
        plot_kwargs['title'] = kwargs.get('title', n_x * ('', ))
        plot_kwargs['legend'] = kwargs.get('legend', True)
        plot_kwargs['kind'] = 'scatter'

        solution.plot(*plots, x_data=ext_data_x, y_data=ext_data_y, **plot_kwargs)

    def plot_1d(
            self,
            quantiles: Optional[tuple[Numeric, Numeric]] = None,
            resolution: Optional[int] = 200,
            backend: Optional[str] = None,
            **kwargs
    ) -> None:
        """

        :param quantiles:
        :param resolution:
        :param backend:
        :param kwargs:
        :return:
        """
        X_train = self._X_train.values
        y_train = self._y_train.values
        n_x = X_train.shape[0]

        if n_x > 1:
            raise RuntimeError("The method plot_1d is only supported for Gaussian processes with one feature.")

        x_min = X_train.min()
        x_max = X_train.max()
        x_min, x_max = x_min - 0.25 * (x_max - x_min), x_max + 0.25 * (x_max - x_min)
        X = np.linspace([float(x_min)], [float(x_max)], resolution, axis=1)

        y, var = self.predict(X, noise_free=True)
        lb, ub = self.predict_quantiles(quantiles=quantiles, mean=y, var=var)

        solution = _GPSeries(backend=backend)
        if backend is None:
            backend = solution.plot_backend

        names = ['x', 'y']
        vector = {
            'x': {
                'values_or_names': self._features,
                'description': [''],
                'labels': self._features,
                'units': [''],
                'shape': (1, 0),
                'data_format': ca.DM
            },
            'y': {
                'values_or_names': ['mean'],
                'description': [''],
                'labels': self._labels,
                'units': [''],
                'shape': (1, 0),
                'data_format': ca.DM
            }
        }
        solution.setup(*names, **vector)

        solution.set('x', X)
        solution.set('y', y)

        ext_data_x = [{'data': X_train, 'subplot': 0, 'label': 'data'}]
        ext_data_y = [{'data': y_train, 'subplot': 0, 'label': 'data', 'kind': 'scatter'}]

        plot_kwargs = kwargs.copy()
        if backend == 'bokeh':
            if kwargs.get("output_notebook", False):
                plot_kwargs["figsize"] = kwargs.get("figsize", (300, 300))
            else:
                plot_kwargs["figsize"] = kwargs.get("figsize", (500, 500))

            plot_kwargs["line_width"] = kwargs.get("line_width", 2)
            plot_kwargs["major_label_text_font_size"] = kwargs.get("major_label_text_font_size", "12pt")
            plot_kwargs["axis_label_text_font_size"] = kwargs.get("axis_label_text_font_size", "12pt")
        plot_kwargs['color'] = kwargs.get('color', ['#3465a4', 'k'])
        plot_kwargs['marker'] = kwargs.get('marker', 'o')
        if backend == 'bokeh':
            plot_kwargs['marker_size'] = kwargs.get('marker_size', 15)
        elif backend == 'matplotlib':
            plot_kwargs['marker_size'] = kwargs.get('marker_size', 60)
        plot_kwargs['fill_between'] = [
            {'x': self._features[0], 'lb': lb, 'ub': ub, 'label': 'confidence', 'line_color': '#204a87',
             'line_width': .5, 'fill_color': '#729fcf', 'fill_alpha': .2}
        ]
        plot_kwargs['title'] = kwargs.get('title', "GP regression")
        plot_kwargs['legend'] = kwargs.get('legend', True)

        solution.plot((self._features[0], 'mean'), x_data=ext_data_x, y_data=ext_data_y, **plot_kwargs)

    def plot_prediction_error(
            self,
            X_query: Array,
            y_query: Array,
            noise_free: bool = False,
            backend: Optional[str] = None,
            **kwargs
    ) -> None:
        """

        :param X_query:
        :param y_query:
        :param noise_free:
        :param backend:
        :param kwargs:
        :return:
        """
        post_mean, post_var = self.predict(X_query, noise_free=noise_free)
        error = np.absolute(post_mean - y_query)
        post_std = np.sqrt(post_var)
        n_samples = X_query.shape[1]

        solution = TimeSeries(backend=backend)
        if backend is None:
            backend = solution.plot_backend

        names = ['t', 'x']
        vector = {
            't': {
                'values_or_names': ['t'],
                'description': [''],
                'labels': [''],
                'units': [''],
                'shape': (1, 0),
                'data_format': ca.DM
            },
            'x': {
                'values_or_names': ['error', 'post_std'],
                'description': 2 * [''],
                'labels': 2 * [''],
                'units': 2 * [''],
                'shape': (2, 0),
                'data_format': ca.DM
            }
        }
        solution.setup(*names, **vector)

        solution.set('t', np.linspace([0.], [n_samples - 1], n_samples, axis=1))
        solution.set('x', np.append(error, post_std, axis=0))

        plot_kwargs = kwargs.copy()
        plot_kwargs['title'] = ('prediction error', 'prediction function standard deviation')
        plot_kwargs['xlabel'] = ('', '')
        plot_kwargs['ylabel'] = ('', '')
        plot_kwargs['kind'] = 'scatter'
        plot_kwargs['marker'] = 'o'
        plot_kwargs['marker_size'] = 5 if backend == 'bokeh' else 20

        solution.plot(('t', 'error'), ('t', 'post_std'), **plot_kwargs)


class GPArray:
    """"""
    def __init__(self, n_gps: int) -> None:
        """Constructor method"""
        self._n_gps = n_gps
        self._gps = np.empty((self._n_gps, 1), dtype=object)

    __array_ufunc__ = None  # https://stackoverflow.com/questions/38229953/array-and-rmul-operator-in-python-numpy

    def __len__(self) -> int:
        """Length method"""
        return self._n_gps

    def __iter__(self) -> Iterator["GaussianProcess"]:
        """Item iteration method"""
        for k in range(self._n_gps):
            gp = self._gps[k, 0]
            if gp is None:
                gp = object.__new__(GaussianProcess)
                self._gps[k, 0] = gp
            yield gp

    def __getitem__(self, item: int) -> Optional[GaussianProcess]:
        """Item getter method"""
        return self._gps[item, 0]

    def __rmatmul__(self, other: Array) -> GPArray:
        """Multiplication method (from the right side)"""
        other = convert(other, ca.DM)
        features = []
        labels = ca.SX.zeros(self._n_gps)
        for k in range(self._n_gps):
            gp = self._gps[k, 0]

            query = [ca.SX.sym(k) for k in gp.features]
            out, _ = gp.predict(ca.vertcat(*query))

            features += query
            labels[k] += out
        # NOTE: @-operator doesn't seem to work because we set __array_ufunc__ to None
        equations = Equations(ca.SX, {'matmul': ca.mtimes(other, labels)})
        return equations.to_function(f'{type(GaussianProcess).__name__}_matmul', *features)

    # def append(self, gp: GaussianProcess) -> None:
    #     """
    #
    #     :param gp:
    #     :return:
    #     """
    #
    # def pop(self) -> GaussianProcess:
    #     """
    #
    #     :return:
    #     """
    #
    # def remove(self, item) -> None:
    #     """
    #
    #     :param item:
    #     :return:
    #     """


__all__ = [
    'GaussianProcess',
    'GPArray'
]



#
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2025 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import annotations


import asyncio
from typing import Any, Dict, List, Optional, Sequence

from .opcua_async import AsyncOPCUAClient, IOMapping


class OPCUASimpleControlLoop:
    """OPC UA control loop (controller+estimator) interface similar to SimpleControlLoop.

    Instead of simulating a local plant model, this loop reads the current state (and
    optionally parameters/measurements) from an OPC UA server, invokes a controller (optionally an estimator)
    and writes the computed control inputs back to the server.

    Design goals
    ------------
    - Mirror the convenience of SimpleControlLoop: automatic controller setup, flexible
      support for MPC/OCP, PID, learning-based and generic controllers.
    - Keep plant-agnostic: we do not advance dynamics; we only exchange data with the server.
    - Allow an optional estimator object whose update/estimate/predict method is called
      after each control action (e.g. for filtering noisy measurements).
    - Provide safe shutdown: write zeros (or user-specified values) to all control outputs
      when the loop exits.

    Minimal example
    ---------------

        mapping = IOMapping(
            reads={
                "theta": {"node": f"ns={ns};s=Pendulum/Angle_rad"},
            },
            writes={
                "u": {"node": f"ns={ns};s=Pendulum/Torque_Nm"},
            },
        )
        loop = OPCUASimpleControlLoop(
            endpoint=endpoint,
            mapping=mapping,
            controller=nmpc,
            state_aliases=["theta"],
            control_aliases=["u"],
            period=0.02,
        )
        await loop.run(max_iters=1000)

    Controller interface detection (similar to SimpleControlLoop):
    - MPC / OCP: must have .optimize(x, cp=None, **kwargs); we pass the current state subset and optional params.
    - ANN / learning-based: must have .predict(x) returning control vector.
    - PID controllers: must have .call(pv=x_current).
    - Generic controller: fallback to .call(x=<state>, p=<params>)

    Estimator (optional):
    - If has .estimate(), call it after writing control.
    - Else if has .predict(), call it.

    Notes
    -----
    - The ordering of state_aliases defines the vector passed to the controller.
    - If parameter_aliases are provided, their ordering defines cp for optimize().
    - Safe shutdown values can be overridden by passing a dict mapping control_alias to value.
    """
    def __init__(
        self,
        endpoint: str,
        mapping: IOMapping,
        controller: Any,
        estimator: Optional[Any] = None,
        state_aliases: Sequence[str] | None = None,
        control_aliases: Sequence[str] | None = None,
        parameter_aliases: Sequence[str] | None = None,
        period: float = 0.05,
        reconnect_backoff: tuple[float, float] = (0.5, 5.0),
        safe_shutdown: Optional[Dict[str, float]] = None,
        # AsyncOPCUAClient connection/security options (passthrough)
        timeout: float = 2.0,
        security_mode: Optional[str] = None,
        security_policy: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        cert_path: Optional[str] = None,
        key_path: Optional[str] = None,
        max_reconnect_attempts: Optional[int] = None,
    ) -> None:
        """Initialize the OPC UA control loop.

        Parameters
        ----------
        endpoint : str
            OPC UA server endpoint URL, e.g., "opc.tcp://127.0.0.1:4840/server/".
        mapping : IOMapping
            Mapping between aliases and OPC UA node IDs for reads and writes.
        controller : Any
            Controller object (NMPC, LMPC, OCP, PID, or machine learning model) with optimize/predict/call interface.
        estimator : Optional[Any], (MHE, KF, UKF or machine learning model) default=None
            Optional estimator object with estimate() or predict() methods.
        state_aliases : Sequence[str] | None, default=None
            List of state variable aliases to read and pass to controller. If None, uses all read aliases.
        control_aliases : Sequence[str] | None, default=None
            List of control variable aliases to write. If None, uses all write aliases.
        parameter_aliases : Sequence[str] | None, default=None
            List of parameter aliases to read and pass to controller as cp argument.
        period : float, default=0.05
            Control loop sampling period in seconds.
        reconnect_backoff : tuple[float, float], default=(0.5, 5.0)
            Exponential backoff bounds (min, max) in seconds for reconnection attempts.
        safe_shutdown : Optional[Dict[str, float]], default=None
            Dictionary mapping control aliases to safe shutdown values. If None, all controls set to 0.0.
        timeout : float, default=2.0
            Timeout in seconds for OPC UA read/write operations.
        security_mode : Optional[str], default=None
            OPC UA security mode: "None", "Sign", or "SignAndEncrypt".
        security_policy : Optional[str], default=None
            OPC UA security policy: "Basic256Sha256", "Aes128_Sha256_RsaOaep", etc.
        username : Optional[str], default=None
            Username for OPC UA authentication.
        password : Optional[str], default=None
            Password for OPC UA authentication.
        cert_path : Optional[str], default=None
            Path to client certificate file (DER or PEM format).
        key_path : Optional[str], default=None
            Path to client private key file (PEM format).
        max_reconnect_attempts : Optional[int], default=None
            Maximum number of consecutive reconnection attempts. If None, retry indefinitely.
        """
        self.endpoint = endpoint
        self.mapping = mapping
        self.controller = controller
        self.estimator = estimator
        self.period = float(period)
        self.state_aliases = list(state_aliases or mapping.reads.keys())
        self.control_aliases = list(control_aliases or mapping.writes.keys())
        self.parameter_aliases = list(parameter_aliases or [])
        self.client = AsyncOPCUAClient(
            endpoint=endpoint,
            mapping=mapping,
            timeout=timeout,
            security_mode=security_mode,
            security_policy=security_policy,
            username=username,
            password=password,
            cert_path=cert_path,
            key_path=key_path,
            reconnect_backoff=reconnect_backoff,
            max_reconnect_attempts=max_reconnect_attempts,
        )
        self.safe_shutdown = (
            dict(safe_shutdown)
            if safe_shutdown is not None
            else {alias: 0.0 for alias in self.control_aliases}
        )

        # Feature flags (similar to SimpleControlLoop)
        self._is_mpc = getattr(controller, "type", None) in {"NMPC", "LMPC"}
        self._is_ocp = getattr(controller, "type", None) == "OCP"
        self._is_pid = getattr(controller, "type", None) == "PID"
        self._has_optimize = hasattr(controller, "optimize") and (self._is_mpc or self._is_ocp)
        self._has_predict = hasattr(controller, "predict")
        self._has_call = hasattr(controller, "call")

        self._est_has_estimate = hasattr(estimator, "estimate") if estimator else False
        self._est_has_predict = hasattr(estimator, "predict") if estimator else False

        # Ensure setup() executed if available
        if hasattr(controller, "is_setup") and not controller.is_setup():
            controller.setup()
        if estimator and hasattr(estimator, "is_setup") and not estimator.is_setup():
            estimator.setup()

    async def _read_values(self) -> Dict[str, float]:
        """Read state and parameter values from the OPC UA server.

        Returns
        -------
        Dict[str, float]
            Dictionary mapping aliases to their float values. Non-numeric values are skipped.
        """
        data = await self.client.read(self.state_aliases + self.parameter_aliases)
        out: Dict[str, float] = {}
        for k, v in data.items():
            try:
                out[k] = float(v["value"])
            except Exception:
                # Non-numeric -> skip
                pass
        return out

    def _build_vectors(self, values: Dict[str, float]) -> tuple[List[float], Optional[List[float]]]:
        """Build state and parameter vectors from raw values dictionary.

        Parameters
        ----------
        values : Dict[str, float]
            Dictionary of alias-value pairs read from the server.

        Returns
        -------
        tuple[List[float], Optional[List[float]]]
            Tuple containing:
            - x_vec: State vector ordered according to state_aliases
            - cp_vec: Parameter vector ordered according to parameter_aliases, or None if no parameters
        """
        x_vec = [values.get(a, 0.0) for a in self.state_aliases]
        cp_vec = [values.get(a, 0.0) for a in self.parameter_aliases] if self.parameter_aliases else None
        return x_vec, cp_vec

    def _compute_control(self, x_vec: List[float], cp_vec: Optional[List[float]]) -> List[float]:
        """Compute control action using the controller.

        Parameters
        ----------
        x_vec : List[float]
            Current state vector.
        cp_vec : Optional[List[float]]
            Optional parameter vector for the controller.

        Returns
        -------
        List[float]
            Computed control action vector.

        Raises
        ------
        RuntimeError
            If the controller does not expose a recognized interface (optimize/predict/call).
        """
        def _tolist(u: Any) -> List[float]:
            """Assume controller returns CasADi DM/MX; convert via toarray() and flatten.
            
            Parameters
            ----------
            u : Any
                Controller output (expected to be CasADi DM/MX or array-like).
            
            Returns
            -------
            List[float]
                Flattened list of control values.
            """
            arr = u.toarray()  # Expect DM/MX; if not, this will raise and reveal misuse early.
            return [float(v) for v in arr.flatten()]

        if self._has_optimize:
            return _tolist(self.controller.optimize(x_vec, cp=cp_vec))
        if self._has_predict:
            return _tolist(self.controller.predict(x_vec))
        if self._is_pid and self._has_call:
            return _tolist(self.controller.call(pv=x_vec))
        if self._has_call:
            return _tolist(self.controller.call(x=x_vec, p=cp_vec))
        raise RuntimeError("Controller object does not expose optimize/predict/call interface")

    async def _estimator_step(self) -> None:
        """Execute one estimator step if an estimator is configured.

        Calls estimate() if available, otherwise predict(). Errors are silently caught
        for robustness, as estimator failures should not crash the control loop.
        """
        if not self.estimator:
            return
        try:
            if self._est_has_estimate:
                self.estimator.estimate()
            elif self._est_has_predict:
                self.estimator.predict()
        except Exception:
            # Non-critical; ignore estimator errors for robustness
            pass

    async def run(self, max_iters: Optional[int] = None) -> None:
        """Run the control loop asynchronously.

        Executes the sense-compute-actuate cycle at the configured period until interrupted
        or max_iters is reached. On exit, writes safe shutdown values to all control outputs.

        Parameters
        ----------
        max_iters : Optional[int], default=None
            Maximum number of iterations to run. If None, runs indefinitely until interrupted.

        Notes
        -----
        The control loop performs the following steps each iteration:
        1. Read state and parameter values from OPC UA server
        2. Compute control action using the controller
        3. Write control action to OPC UA server
        4. Execute estimator step (if configured)
        5. Sleep for the configured period

        On exit (normal or via exception), safe shutdown values are written to ensure
        the system is left in a safe state.
        """
        await self.client.connect()
        try:
            k = 0
            while True:
                values = await self._read_values()
                x_vec, cp_vec = self._build_vectors(values)
                u_vec = self._compute_control(x_vec, cp_vec)
                out: Dict[str, float] = {}
                for i, alias in enumerate(self.control_aliases):
                    out[alias] = float(u_vec[i]) if i < len(u_vec) else 0.0
                await self.client.write(out)
                await self._estimator_step()
                await asyncio.sleep(self.period)
                if max_iters is not None:
                    k += 1
                    if k >= max_iters:
                        break
        finally:
            try:
                if self.safe_shutdown:
                    await self.client.write(self.safe_shutdown)
            except Exception:
                pass
            await self.client.disconnect()

    def run_sync(self, max_iters: Optional[int] = None) -> None:
        """Synchronous convenience wrapper around the async run method.

        Parameters
        ----------
        max_iters : Optional[int], default=None
            Maximum number of iterations to run. If None, runs indefinitely until interrupted.

        Notes
        -----
        This method creates a new event loop, runs the async control loop, and cleans up.
        Use this method when calling from synchronous code. For async contexts, use run() directly.
        """
        asyncio.run(self.run(max_iters=max_iters))



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from abc import ABCMeta, abstractmethod
from typing import Any, Optional, Sequence, TypeVar, Union

import casadi as ca
import numpy as np
from scipy.special import gammaln


Numeric = Union[int, float]
Pr = TypeVar('Pr', bound='Prior')


class Distribution(metaclass=ABCMeta):
    """"""
    def __init__(self):
        """Constructor method"""
        self._log_function = None
        self._initialize()

    def __call__(self, *args, **kwargs) -> Union[ca.SX, ca.MX, ca.DM]:
        """Calling method"""
        log = kwargs.get('log')
        if log is None:
            log = False

        if log:
            return self._log_function(*args)

    @abstractmethod
    def _initialize(self) -> None:
        """

        :return:
        """
        pass


class Gaussian(Distribution):
    """"""
    def _initialize(self) -> None:
        """

        :return:
        """
        # TODO: Multivariate extension
        y = ca.SX.sym('y')
        mu = ca.SX.sym('mu')
        var = ca.SX.sym('var')

        log_pdf = -(y - mu) ** 2. / (2. * var) - ca.log(2. * ca.pi * var) / 2.

        self._log_function = ca.Function('log_pdf',
                                         [y, mu, var],
                                         [log_pdf])


class Laplace(Distribution):
    """"""
    def _initialize(self) -> None:
        """

        :return:
        """
        # TODO: Multivariate extension
        y = ca.SX.sym('y')
        mu = ca.SX.sym('mu')
        var = ca.SX.sym('var')

        b = ca.sqrt(var / 2.)
        log_pdf = -ca.fabs(y - mu) / b - ca.log(2. * b)

        self._log_function = ca.Function('log_pdf',
                                         [y, mu, var],
                                         [log_pdf])


class StudentsT(Distribution):
    """"""
    def _initialize(self) -> None:
        """

        :return:
        """
        # TODO: Multivariate extension
        y = ca.SX.sym('y')
        mu = ca.SX.sym('mu')
        var = ca.SX.sym('var')
        nu = ca.SX.sym('nu')

        log_Z = gammaln((nu + 1.) / 2.) - gammaln(nu / 2.) - ca.log(var * (nu - 2.) * ca.pi) / 2.
        log_pdf = log_Z - (nu + 1.) / 2. * ca.log(1. + (y - mu) ** 2. / (var * (nu - 2.)))

        self._log_function = ca.Function('log_pdf',
                                         [y, mu, var, nu],
                                         [log_pdf])


class Prior(metaclass=ABCMeta):
    """"""
    def __init__(self, name: str) -> None:
        """Constructor method"""
        self._name = name
        self._pdf = None

    def __call__(self, *args, **kwargs) -> Union[ca.SX, ca.MX, ca.DM]:
        """Calling method"""
        x = args[0]
        p = self._get_parameter_values()
        return self._pdf(x, *p, **kwargs)

    @staticmethod
    def _check_dimensionality(parameter: Any, name: str) -> Numeric:
        """

        :param parameter:
        :param name:
        :return:
        """
        if parameter is not None:
            if not isinstance(parameter, (int, float)):
                if isinstance(parameter, (list, tuple, np.ndarray)):
                    if len(parameter) == 1:
                        parameter = parameter[0]
                    else:
                        raise ValueError(f"Dimension mismatch. Expected 1, got {len(parameter)}.")
                else:
                    raise TypeError(f"Unexpected type '{type(parameter).__name__}' for keyword argument '{name}'")

        return parameter

    @abstractmethod
    def _get_parameter_values(self) -> (Union[Numeric, Sequence[Numeric]], ...):
        """

        :return:
        """
        pass

    @property
    def name(self) -> str:
        """

        :return:
        """
        return self._name

    @staticmethod
    def gaussian(
            mean: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            variance: Optional[Union[Numeric, Sequence[Numeric]]] = None
    ) -> Pr:
        """

        :param mean:
        :param variance:
        :return:
        """
        return GaussianPrior(mean=mean, variance=variance)

    @staticmethod
    def laplace(
            mean: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            variance: Optional[Union[Numeric, Sequence[Numeric]]] = None
    ) -> Pr:
        """

        :param mean:
        :param variance:
        :return:
        """
        return LaplacePrior(mean=mean, variance=variance)

    @staticmethod
    def students_t(
            mean: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            variance: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            nu: Optional[Union[Numeric, Sequence[Numeric]]] = None
    ) -> Pr:
        """

        :param mean:
        :param variance:
        :param nu:
        :return:
        """
        return StudentsTPrior(mean=mean, variance=variance, nu=nu)

    @staticmethod
    def delta() -> Pr:
        """

        :return:
        """
        return DeltaPrior()


class _MeanVariancePrior(Prior):
    """"""
    def __init__(
            self,
            name: str,
            mean: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            variance: Optional[Union[Numeric, Sequence[Numeric]]] = None
    ) -> None:
        """Constructor method"""
        super().__init__(name)

        self._mean = self._check_dimensionality(mean, 'mean')
        self._variance = self._check_dimensionality(variance, 'variance')

    def _get_parameter_values(self) -> (Union[Numeric, Sequence[Numeric]], ...):
        """

        :return:
        """
        return self._mean, self._variance

    @property
    def mean(self) -> Optional[Numeric]:
        """

        :return:
        """
        return self._mean

    @mean.setter
    def mean(self, value: Numeric) -> None:
        self._mean = self._check_dimensionality(value, 'mean')

    @property
    def variance(self) -> Optional[Numeric]:
        """

        :return:
        """
        return self._variance

    @variance.setter
    def variance(self, value: Numeric) -> None:
        self._variance = self._check_dimensionality(value, 'variance')


class GaussianPrior(_MeanVariancePrior):
    """"""
    def __init__(
            self,
            mean: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            variance: Optional[Union[Numeric, Sequence[Numeric]]] = None
    ) -> None:
        """Constructor method"""
        super().__init__('Gaussian', mean=mean, variance=variance)

        self._pdf = Gaussian()


class LaplacePrior(_MeanVariancePrior):
    """"""
    def __init__(
            self,
            mean: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            variance: Optional[Union[Numeric, Sequence[Numeric]]] = None
    ) -> None:
        """Constructor method"""
        super().__init__('Laplace', mean=mean, variance=variance)

        self._pdf = Laplace()


class StudentsTPrior(_MeanVariancePrior):
    """"""
    def __init__(
            self,
            mean: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            variance: Optional[Union[Numeric, Sequence[Numeric]]] = None,
            nu: Optional[Union[Numeric, Sequence[Numeric]]] = None
    ) -> None:
        """Constructor method"""
        super().__init__('Students_T', mean=mean, variance=variance)

        self._nu = self._check_dimensionality(nu, 'nu')
        self._pdf = StudentsT()

    def _get_parameter_values(self) -> (Union[Numeric, Sequence[Numeric]], ...):
        """

        :return:
        """
        return super()._get_parameter_values(), self._nu

    @property
    def nu(self) -> Optional[Numeric]:
        """

        :return:
        """
        return self._nu

    @nu.setter
    def nu(self, value: Numeric) -> None:
        self._nu = self._check_dimensionality(value, 'nu')


class DeltaPrior(Prior):
    """"""
    def __init__(self):
        """Constructor method"""
        super().__init__('Delta')

    def _get_parameter_values(self) -> (Union[Numeric, Sequence[Numeric]], ...):
        """

        :return:
        """
        pass


__all__ = [
    'Prior',
    'GaussianPrior',
    'LaplacePrior',
    'StudentsTPrior',
    'DeltaPrior'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

import warnings

from ....util.util import is_list_like


class Layer:
    """Base layer"""
    def __init__(self, nodes, activation=None, initializer=None, parent=None, **kwargs):
        """Constructor method"""
        self._nodes = nodes
        if activation is None:
            activation = 'sigmoid'
        activation = activation.lower().replace(' ', '_')
        self._activation = activation
        self._set_initializer(initializer, **kwargs)
        self._parent = parent
        self._type = None

    def __len__(self):
        """Length method"""
        return self._nodes

    def _set_initializer(self, initializer, **kwargs):
        """

        :param initializer:
        :param kwargs:
        :return:
        """
        if initializer is not None:
            initializer = initializer.lower().replace(' ', '_')
            if initializer == 'uniform':
                min_val = kwargs.get('min_val')
                max_val = kwargs.get('max_val')
                if min_val is None:
                    min_val = -.05
                if max_val is None:
                    max_val = .05
                initializer = {'uniform': {'a': min_val, 'b': max_val}}
            elif initializer == 'normal':
                mean = kwargs.get('mean')
                std = kwargs.get('std')
                if mean is None:
                    mean = 0.
                if std is None:
                    std = .05
                initializer = {'normal': {'mean': mean, 'std': std}}
            elif initializer == 'constant':
                val = kwargs.get('val')
                if val is None:
                    val = 0
                initializer = {'constant': val}
            elif initializer == 'ones':
                pass
            elif initializer == 'zeros':
                pass
            elif initializer == 'eye':
                pass
            # Dirac initializer will be skipped, since it's only applicable to {3,4,5}-dimensional input tensors
            elif initializer in ['glorot_uniform', 'xavier_uniform']:
                initializer = 'xavier_uniform'
            elif initializer in ['glorot_normal', 'xavier_normal']:
                initializer = 'xavier_normal'
            elif initializer in ['kaiming_uniform', 'he_uniform']:
                if self._activation in ['relu', 'leaky_relu']:
                    initializer = {'kaiming_uniform': {}}

                    if self._activation == 'leaky_relu':
                        a = kwargs.get('a')
                        if a is None:
                            a = 0
                        initializer['kaiming_uniform']['a'] = a

                    mode = kwargs.get('mode')
                    if mode is None:
                        mode = 'fan_in'
                    initializer['kaiming_uniform']['mode'] = mode

                    initializer['kaiming_uniform']['nonlinearity'] = self._activation
                else:
                    initializer = None
                    warnings.warn("It is recommended to use Kaiming uniform weight initialization only with ReLU and "
                                  "leaky ReLU")
            elif initializer in ['kaiming_normal', 'he_normal']:
                if self._activation in ['relu', 'leaky_relu']:
                    initializer = {'kaiming_normal': {}}

                    if self._activation == 'leaky_relu':
                        a = kwargs.get('a')
                        if a is None:
                            a = 0
                        initializer['kaiming_normal']['a'] = a

                    mode = kwargs.get('mode')
                    if mode is None:
                        mode = 'fan_in'
                    initializer['kaiming_normal']['mode'] = mode

                    initializer['kaiming_normal']['nonlinearity'] = self._activation
                else:
                    initializer = None
            elif initializer == 'orthogonal':
                gain = kwargs.get('gain')
                if gain is None:
                    gain = 1
                initializer = {'orthogonal': {'gain': gain}}
            elif initializer == 'sparse':
                sparsity = kwargs.get('sparsity')
                if sparsity is None:
                    sparsity = .1
                std = kwargs.get('std')
                if std is None:
                    std = .01
                initializer = {'sparse': {'sparsity': sparsity, 'std': std}}
            else:
                raise ValueError("Initializer not recognized")
        self._initializer = initializer

    @property
    def activation(self):
        """

        :return:
        """
        return self._activation

    @activation.setter
    def activation(self, arg):
        self._activation = arg

    @property
    def parent(self):
        """

        :return:
        """
        return self._parent

    @parent.setter
    def parent(self, arg):
        self._parent = arg

    @property
    def type(self):
        """

        :return:
        """
        return self._type

    @property
    def initializer(self):
        """

        :return:
        """
        if isinstance(self._initializer, str):
            return self._initializer
        elif isinstance(self._initializer, dict):
            return list(self._initializer)[0]
        return None

    # @initializer.setter
    def set_initializer(self, initializer, **kwargs):
        """

        :param initializer:
        :param kwargs:
        :return:
        """
        self._set_initializer(initializer, **kwargs)

    @staticmethod
    def dense(nodes, activation='linear', initializer=None, dropout=None, where=None, parent=None, **kwargs):
        """

        :param nodes:
        :param activation:
        :param initializer:
        :param dropout:
        :param where:
        :param parent:
        :param kwargs:
        :return:
        """
        if is_list_like(nodes):
            n_layers = len(nodes)

            if is_list_like(activation):
                if len(activation) != n_layers:
                    raise ValueError(f"Dimension mismatch between supplied nodes list of length {n_layers} "
                                     f"and supplied activation function list of length {len(activation)}")
            else:
                activation = n_layers * [activation]

            if is_list_like(initializer):
                if len(initializer) != n_layers:
                    raise ValueError(f"Dimension mismatch between supplied nodes list of length {n_layers} "
                                     f"and supplied initializer list of length {len(initializer)}")
            else:
                initializer = n_layers * [initializer]

            if dropout is not None:
                if where is None:
                    where = 'after'
                else:
                    where = where.lower()

                if is_list_like(dropout):
                    if where in ['before', 'after']:
                        if len(dropout) != n_layers:
                            raise ValueError(f"Dimension mismatch between supplied nodes list of length {n_layers} "
                                             f"and supplied dropout list of length {len(dropout)}")
                    else:
                        raise KeyError(f"Wrong keyword argument '{where}' for keyword 'where'")
                else:
                    dropout = n_layers * [dropout]

                layers = []
                for k in range(n_layers):
                    if where == 'after':
                        layers.append(Dense(nodes[k], activation=activation[k], initializer=initializer[k],
                                            parent=parent, **kwargs))
                        layers.append(Dropout(dropout[k], parent=parent))
                    else:
                        layers.append(Dropout(dropout[k], parent=parent))
                        layers.append(Dense(nodes[k], activation=activation[k], initializer=initializer[k],
                                            parent=parent, **kwargs))
            else:
                layers = [Dense(nodes[k], activation=activation[k], initializer=initializer[k], parent=parent, **kwargs)
                          for k in range(n_layers)]

            return layers
        else:
            if is_list_like(activation):
                raise ValueError(f"Dimension mismatch between supplied nodes and supplied activation function list of "
                                 f"length {len(activation)}")

            if is_list_like(initializer):
                raise ValueError(f"Dimension mismatch between supplied nodes and supplied initializer list of length "
                                 f"{len(initializer)}")

            if dropout is not None:
                if where is None:
                    where = 'after'
                else:
                    where = where.lower()

                if is_list_like(dropout):
                    raise ValueError(f"Dimension mismatch between supplied nodes and supplied activation function list"
                                     f" of length {len(dropout)}")

                if where == 'after':
                    return [Dense(nodes, activation=activation, initializer=initializer, parent=parent, **kwargs),
                            Dropout(dropout, parent=parent)]
                elif where == 'before':
                    return [Dropout(dropout, parent=parent),
                            Dense(nodes, activation=activation, initializer=initializer, parent=parent, **kwargs)]
                else:
                    raise KeyError(f"Wrong keyword argument '{where}' for keyword 'where'")
            else:
                return Dense(nodes, activation=activation, initializer=initializer, parent=parent, **kwargs)

    @staticmethod
    def dropout(rate, parent=None):
        """

        :param rate:
        :param parent:
        :return:
        """
        return Dropout(rate, parent=parent)


class Dense(Layer):
    """Dense layer"""
    def __init__(self, nodes, activation='linear', initializer=None, parent=None, **kwargs):
        """Constructor method"""
        super().__init__(nodes, activation=activation, initializer=initializer, parent=parent, **kwargs)

        self._type = 'dense'

    @property
    def nodes(self):
        """

        :return:
        """
        return self._nodes

    @nodes.setter
    def nodes(self, arg):
        self._nodes = arg


class Dropout(Layer):
    """Dropout layer

    :param rate: Dropout rate
    :type rate: float
    """
    def __init__(self, rate, parent=None):
        """Constructor method"""
        super().__init__(0, parent=parent)
        self._activation = None
        self._type = 'dropout'
        self._rate = rate

    @property
    def rate(self):
        """

        :return:
        """
        return self._rate

    @rate.setter
    def rate(self, arg):
        self._rate = arg

    def set_initializer(self, initializer, **kwargs):
        """

        :param initializer:
        :param kwargs:
        :return:
        """
        print("Initialization is not required for Dropout layers")


__all__ = [
    'Layer',
    'Dense',
    'Dropout'
]



#
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2025 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Any, Iterable, Optional, Tuple, Sequence, TypedDict, Union

# Optional dependency: asyncua (opcua-asyncio)
try:
    from asyncua import Client as UAClient
    from asyncua import ua
except Exception:  # pragma: no cover - handled at runtime with clear error
    UAClient = None  # type: ignore
    ua = None  # type: ignore


class ReadItem(TypedDict, total=False):
    """Structured item returned by reads.

    Keys:
    - value: The decoded value (possibly scaled to engineering units)
    - status: UA status code object (if asyncua available) or None
    - ts: Server/source timestamp if available; None in minimal version
    """
    value: Any
    status: Any
    ts: Any


Number = Union[int, float]


@dataclass
class IOMapping:
    """
    A simple dictionary of nicknames (aliases) to real OPC UA addresses (NodeIds).

    Why aliases? Because typing a long NodeId like "ns=2;s=Plant/Speed" everywhere is hard to read
    and easy to mistype. So we map:
        alias -> NodeId string

    We also store two optional things per alias:
    - scale: clever way to convert raw numbers to pretty numbers (engineering units).
             If scale is {a: 2, b: 3}, then pretty = 2*raw + 3.
    - bounds: minimum and maximum allowed values when we write, to keep things safe.

    Example mapping (YAML-like):
        reads:
          y_speed: { node: 'ns=2;s=Plant/Speed',  scale: {a: 1.0, b: 0.0} }
        writes:
          u_pitch: { node: 'ns=2;s=Plant/BladePitch_deg', bounds: [0, 30] }
    """

    reads: Dict[str, Dict[str, Any]] = field(default_factory=dict)
    writes: Dict[str, Dict[str, Any]] = field(default_factory=dict)

    def node_of(self, key: str) -> str:
        """Return the NodeId string for an alias.

        Parameters
        ----------
        key: str
            Alias as defined in reads/writes.

        Returns
        -------
        str
            The configured NodeId string, e.g. "ns=2;s=Plant/Speed".

        Raises
        ------
        KeyError
            If the alias is not present in either reads or writes.
        """
        if key in self.reads:
            return self.reads[key]["node"]
        if key in self.writes:
            return self.writes[key]["node"]
        raise KeyError(f"Alias '{key}' not found in mapping")

    def get_scale(self, key: str) -> Optional[Tuple[float, float]]:
        """Return (a, b) for linear scale pretty = a*raw + b, if configured.

        Returns None when no scale exists for this alias.
        """
        cfg = self.reads.get(key) or self.writes.get(key)
        if not cfg:
            return None
        scale = cfg.get("scale")
        if not scale:
            return None
        a = float(scale.get("a", 1.0))
        b = float(scale.get("b", 0.0))
        return a, b

    def get_bounds(self, key: str) -> Optional[Tuple[float, float]]:
        """Return (min, max) bounds if defined for a write alias; else None."""
        cfg = self.writes.get(key)
        if not cfg:
            return None
        bnd = cfg.get("bounds")
        if not bnd:
            return None
        return float(bnd[0]), float(bnd[1])


def build_mapping_from_model(
    model: Any,
    ns_idx: int,
    base_path: str,
    read_from: Sequence[str] | str = ("y",),
    write_from: Sequence[str] | str = ("u",),
    write_bounds: Optional[Dict[str, Tuple[float, float]]] = None,
    scales: Optional[Dict[str, Tuple[float, float]]] = None,
    node_overrides: Optional[Dict[str, str]] = None,
    include_read_names: Optional[Sequence[str]] = None,
    include_write_names: Optional[Sequence[str]] = None,
) -> IOMapping:
    """Build an IOMapping from a HILO model using variable names and a path template.

    - NodeId format: f"ns={ns_idx};s={base_path}/{name}"
    - read_from: which variable groups to expose as reads: any of 'x','y','z','u','p'.
    - write_from: variable groups to expose as writes (commonly 'u').
    - write_bounds: optional per-name (lo, hi) bounds applied on write.
    - scales: optional per-name (a, b) linear scaling for read/write.

    This makes it nearly zero-boilerplate to connect a controller to an OPC UA server
    when the server exposes nodes named exactly after the model variable names under a
    common base path.
    Parameters
    ----------
    model: Any
        HILO model instance; must expose name lists like `input_names`, `dynamical_state_names`, etc.
    ns_idx: int
        Namespace index where the server publishes the nodes.
    base_path: str
        Base path segment for nodes, e.g. "Pendulum" so nodes become "Pendulum/theta".
    read_from: Sequence[str] | str
        Model groups to expose as readable aliases; subset of {"x","y","z","u","p"}.
    write_from: Sequence[str] | str
        Model groups to expose as writable aliases (commonly {"u"}).
    write_bounds: dict | None
        Optional per-name bounds applied before write, as {name: (lo, hi)}.
    scales: dict | None
        Optional per-name linear scaling (a, b) applied on read and inversely on write.
    node_overrides: dict | None
        Optional explicit NodeId per name; takes precedence over the template.
    include_read_names: Sequence[str] | None
        If provided, limit read aliases to this exact subset of names.
    include_write_names: Sequence[str] | None
        If provided, limit write aliases to this exact subset of names.

    Returns
    -------
    IOMapping
        Mapping ready to use with AsyncOPCUAClient/OPCUAConnector.

    Notes
    -----
    - NodeId format used here is string-based: f"ns={ns_idx};s={base_path}/{name}".
    - This is convenience only; for complex addressing schemes, pass explicit overrides.
    """
    def _names(kind: str) -> Sequence[str]:
        kind = kind.lower()
        if kind == "x":
            return getattr(model, "dynamical_state_names", [])
        if kind == "y":
            return getattr(model, "measurement_names", [])
        if kind == "z":
            return getattr(model, "algebraic_state_names", [])
        if kind == "u":
            return getattr(model, "input_names", [])
        if kind == "p":
            return getattr(model, "parameter_names", [])
        return []

    if isinstance(read_from, str):
        read_from = (read_from,)
    if isinstance(write_from, str):
        write_from = (write_from,)

    reads: Dict[str, Dict[str, Any]] = {}
    writes: Dict[str, Dict[str, Any]] = {}

    overrides = node_overrides or {}

    include_read: Optional[set] = set(include_read_names) if include_read_names else None
    include_write: Optional[set] = set(include_write_names) if include_write_names else None

    for grp in read_from:
        for name in _names(grp):
            if include_read is not None and name not in include_read:
                continue
            node = overrides.get(name, f"ns={ns_idx};s={base_path}/{name}")
            cfg: Dict[str, Any] = {"node": node}
            if scales and name in scales:
                a, b = scales[name]
                cfg["scale"] = {"a": float(a), "b": float(b)}
            reads[name] = cfg

    for grp in write_from:
        for name in _names(grp):
            if include_write is not None and name not in include_write:
                continue
            node = overrides.get(name, f"ns={ns_idx};s={base_path}/{name}")
            cfg2: Dict[str, Any] = {"node": node}
            if write_bounds and name in write_bounds:
                lo, hi = write_bounds[name]
                cfg2["bounds"] = [float(lo), float(hi)]
            if scales and name in scales:
                a, b = scales[name]
                cfg2["scale"] = {"a": float(a), "b": float(b)}
            writes[name] = cfg2

    return IOMapping(reads=reads, writes=writes)


class AsyncOPCUAClient:
    """
    Tiny async OPC UA client that speaks in aliases, not NodeIds.

    Simple mental picture:
    - endpoint is the doorbell address of the machine box, like
      "opc.tcp://127.0.0.1:4840/freeopcua/server/".
    - connect() rings the bell and enters.
    - read([aliases]) asks for the current numbers by their nicknames.
    - write({alias: value}) sends a new number to the machine.
    - disconnect() politely leaves.

    Usage example:
        client = AsyncOPCUAClient(endpoint, mapping)
        await client.connect()
        data = await client.read(["y_speed"])  # {"y_speed": {"value": 123.4, "status": ..., "ts": ...}}
        await client.write({"u_pitch": 5.0})
        await client.disconnect()
    """

    def __init__(
        self,
        endpoint: str,
        mapping: IOMapping,
        timeout: float = 2.0,
        security_mode: Optional[str] = None,
        security_policy: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        cert_path: Optional[str] = None,
        key_path: Optional[str] = None,
        reconnect_backoff: Tuple[float, float] = (0.5, 5.0),
        max_reconnect_attempts: Optional[int] = None,
    ) -> None:
        """Store connection settings and the alias mapping.

        Parameters
        ----------
        endpoint: str
            OPC UA endpoint URL, e.g. "opc.tcp://127.0.0.1:4840/freeopcua/demo/".
        mapping: IOMapping
            Alias-to-NodeId mapping including optional scale and bounds.
        timeout: float
            Request timeout (seconds).
        security_mode, security_policy, username, password, cert_path, key_path
            Optional security/authentication parameters. Not used in the minimal demo.
        reconnect_backoff: (float, float)
            Initial and maximum delays (s) for exponential backoff during reconnect.
        max_reconnect_attempts: int | None
            If set, stop retrying after this many consecutive reconnect attempts.

        Raises
        ------
        ImportError
            If asyncua is not installed in the current environment.
        """
        self._endpoint = endpoint
        self._mapping = mapping
        self._timeout = timeout
        self._security_mode = security_mode
        self._security_policy = security_policy
        self._username = username
        self._password = password
        self._cert_path = cert_path
        self._key_path = key_path
        self._client: Optional[Any] = None
        # Reconnection config/state
        self._reconnect_backoff = reconnect_backoff
        self._max_reconnect_attempts = max_reconnect_attempts
        self._consec_failures = 0

        if UAClient is None:
            raise ImportError(
                "asyncua (opcua-asyncio) is not installed. Install with 'poetry install -E connectivity' or 'pip install asyncua'."
            )

    async def connect(self) -> None:
        """Open a session with the server.

        Applies minimal security/user configuration if provided. For local demos this
        typically uses an unsecured connection.
        """
        client = UAClient(self._endpoint, timeout=self._timeout)  # type: ignore[arg-type]

        # Security configuration (minimal; extend as needed)
        if self._security_mode and self._security_policy:
            # Map simple strings to ua.SecurityPolicyURIs when provided
            policy_uri = getattr(ua.SecurityPolicyURIS, self._security_policy, None) if ua else None
            mode = getattr(ua.MessageSecurityMode, self._security_mode, None) if ua else None
            if policy_uri and mode:
                await client.set_security(policy_uri, certificate=self._cert_path, private_key=self._key_path, mode=mode)  # type: ignore[func-returns-value]

        if self._username and self._password:
            client.set_user(self._username)
            client.set_password(self._password)

        await client.connect()
        self._client = client

    async def disconnect(self) -> None:
        """Close the session when finished."""
        if self._client is not None:
            await self._client.disconnect()
            self._client = None

    def _require_client(self) -> Any:
        """Return the internal asyncua client; require an active connection.

        Raises
        ------
        RuntimeError
            If connect() has not been called successfully yet.
        """
        if self._client is None:
            raise RuntimeError("Client is not connected. Call connect() first.")
        return self._client

    async def ensure_connected(self) -> None:
        """Ensure there is an active session; connect if needed."""
        if self._client is None:
            await self.connect()

    async def _attempt_reconnect(self) -> None:
        """Reconnect with exponential backoff.

        Raises after exceeding max_reconnect_attempts when configured.
        """
        import asyncio
        lo, hi = self._reconnect_backoff
        delay = max(0.0, float(lo))
        attempts = 0
        while True:
            try:
                await self.connect()
                self._consec_failures = 0
                return
            except Exception:
                attempts += 1
                self._consec_failures += 1
                if self._max_reconnect_attempts is not None and attempts >= self._max_reconnect_attempts:
                    raise
                # backoff
                await asyncio.sleep(delay)
                delay = min(delay * 2 if delay > 0 else lo, hi)

    async def read(self, keys: Iterable[str]) -> Dict[str, ReadItem]:
        """Read current values for a list of aliases.

        Steps for each alias:
        - Look up the real NodeId string from the mapping.
        - Ask the server for the value.
        - If a scale is defined, apply pretty = a*raw + b.
        - Return a small dict with value, status, and timestamp (ts is None in this minimal version).
        Parameters
        ----------
        keys: Iterable[str]
            Aliases to read (must be present in mapping.reads or writes).

        Returns
        -------
        dict[str, ReadItem]
            A dictionary keyed by alias with value/status/ts entries.
        """
        await self.ensure_connected()
        client = self._require_client()
        out: Dict[str, ReadItem] = {}
        for k in keys:
            try:
                nodeid = self._mapping.node_of(k)
                node = client.get_node(nodeid)
                value = await node.read_value()
            except Exception:
                # one reconnect attempt then retry once
                await self._attempt_reconnect()
                client = self._require_client()
                node = client.get_node(self._mapping.node_of(k))
                value = await node.read_value()
            status = ua.StatusCode(ua.StatusCodes.Good) if ua else None
            tsrv = None
            # Basic scaling: eng = a*raw + b if provided
            scale = self._mapping.get_scale(k)
            if scale:
                a, b = scale
                try:
                    value = a * float(value) + b
                except Exception:
                    # leave as-is if not numeric
                    pass
            out[k] = {"value": value, "status": status, "ts": tsrv}
        return out

    async def write(self, data: Dict[str, Number]) -> None:
        """Write values by alias.

        Steps for each alias:
        - If scale exists, reverse it: raw = (pretty - b)/a before sending to the server.
        - If bounds exist, clamp the value inside [min, max] for safety.
        - Send the value to the NodeId.
        Parameters
        ----------
        data: dict[str, Number]
            Mapping from write alias to numeric value in engineering units.
        """
        await self.ensure_connected()
        client = self._require_client()
        for k, v in data.items():
            nodeid = self._mapping.node_of(k)
            # Apply inverse scaling if provided: raw = (eng - b)/a
            scale = self._mapping.get_scale(k)
            if scale:
                a, b = scale
                try:
                    v = (float(v) - b) / a
                except Exception:
                    # If scaling fails (e.g., non-numeric value), use original value
                    pass
            bounds = self._mapping.get_bounds(k)
            if bounds:
                lo, hi = bounds
                try:
                    v = min(max(float(v), lo), hi)
                except Exception:
                    pass
            try:
                node = client.get_node(nodeid)
                await node.write_value(v)
            except Exception:
                await self._attempt_reconnect()
                client = self._require_client()
                node = client.get_node(nodeid)
                await node.write_value(v)



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

import shutil
import tempfile


class Session:
    """"""
    def __init__(self) -> None:
        """Constructor method"""
        self._temp_dir = None

    def __enter__(self) -> 'TempDir':
        """Method for entering runtime context"""
        self._temp_dir = TempDir()
        return self._temp_dir

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Method for exiting runtime context"""
        shutil.rmtree(self._temp_dir.path)
        self._temp_dir.close()
        self._temp_dir = None


class TempDir:
    """"""
    def __init__(self) -> None:
        """Constructor method"""
        self._path = tempfile.mkdtemp()

    @property
    def path(self) -> str:
        """
        Path to temporarily created directory

        :return: Path to temporarily created directory
        :rtype: str
        """
        return self._path + '/'

    def close(self) -> None:
        """

        :return:
        """
        self._path = None
